"""
VNTextElement, styledText, textProperty
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

from .base import VNBaseElement
from .styles import VNColor, VNGradient, pathStrokeStyle


@dataclass
class VNTextElement(VNBaseElement):
    """rich text data."""

    string: str
    transform: Optional[List[float]]  # matrix(legacy)
    styledText: Optional[List[singleStyledText]]
    textProperty: Optional[textProperty]


@dataclass
class singleStyledText:
    length: int  # effectiveRange
    fontName: str
    fontSize: float
    alignment: int
    kerning: float
    lineHeight: Optional[Dict]
    fillColor: Optional[VNColor]
    fillGradient: Optional[VNGradient]
    strokeStyle: Optional[pathStrokeStyle]
    strikethrough: bool
    underline: bool

    def convert_text_anchor(self):
        """
        alignment to SVG text-anchor

        0: Left
        1: Center
        2: Right
        3: Justify <- how do i do this?
        """
        return {0: "start", 1: "middle", 2: "end"}.get(self.alignment, "start")


@dataclass
class textProperty:
    # "fixedSize":{"height": float,"width": float}
    # autos don't contain values
    # TODO textProperty is not compatible with format 26 text
    textFrameLimits: Optional[Dict]  # autoWidth, autoHeight, fixedSize
    textFramePivot: Optional[Tuple[float, float]]



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2009 Michel Chatelain.
#               2007 Tavmjong Bah, tavmjong@free.fr
#               2006 Georg Wiora, xorx@quarkbox.de
#               2006 Johan Engelen, johan@shouraizou.nl
#               2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Changes:
#  * This program is derived by Michel Chatelain from funcplot.py.
#    His changes are in the Public Domain.
#  * Michel Chatelain, 17-18 janvier 2009, a partir de funcplot.py
#  * 20 janvier 2009 : adaptation a la version 0.46 a partir de la nouvelle version de funcplot.py
#

import math
import random
from math import pi
from inkex.utils import math_eval

import inkex


def drawfunction(
    t_start,
    t_end,
    xleft,
    xright,
    ybottom,
    ytop,
    samples,
    width,
    height,
    left,
    bottom,
    fx="cos(3*t)",
    fy="sin(5*t)",
    times2pi=False,
    isoscale=True,
    drawaxis=True,
):
    if times2pi:
        t_start *= 2 * pi
        t_end *= 2 * pi

    # coords and scales based on the source rect
    scalex = width / (xright - xleft)
    xoff = left
    coordx = lambda x: (x - xleft) * scalex + xoff  # convert x-value to coordinate
    scaley = height / (ytop - ybottom)
    yoff = bottom
    coordy = lambda y: (ybottom - y) * scaley + yoff  # convert y-value to coordinate

    # Check for isotropic scaling and use smaller of the two scales, correct ranges
    if isoscale:
        if scaley < scalex:
            # compute zero location
            xzero = coordx(0)
            # set scale
            scalex = scaley
            # correct x-offset
            xleft = (left - xzero) / scalex
            xright = (left + width - xzero) / scalex
        else:
            # compute zero location
            yzero = coordy(0)
            # set scale
            scaley = scalex
            # correct x-offset
            ybottom = (yzero - bottom) / scaley
            ytop = (bottom + height - yzero) / scaley

    # functions specified by the user
    f1 = math_eval(fx, "t")
    f2 = math_eval(fy, "t")

    # step is increment of t
    step = (t_end - t_start) / (samples - 1)
    third = step / 3.0
    ds = step * 0.001  # Step used in calculating derivatives

    a = []  # path array
    # add axis
    if drawaxis:
        # check for visibility of x-axis
        if ybottom <= 0 <= ytop:
            # xaxis
            a.append(["M", [left, coordy(0)]])
            a.append(["l", [width, 0]])
        # check for visibility of y-axis
        if xleft <= 0 <= xright:
            # xaxis
            a.append(["M", [coordx(0), bottom]])
            a.append(["l", [0, -height]])

    # initialize functions and derivatives for 0;
    # they are carried over from one iteration to the next, to avoid extra function calculations.
    # print("RET: {}".format(f1(1)))
    x0 = f1(t_start)
    y0 = f2(t_start)

    # numerical derivatives, using 0.001*step as the small differential
    t1 = t_start + ds  # Second point AFTER first point (Good for first point)
    x1 = f1(t1)
    y1 = f2(t1)
    dx0 = (x1 - x0) / ds
    dy0 = (y1 - y0) / ds

    # Start curve
    a.append(["M", [coordx(x0), coordy(y0)]])  # initial moveto
    for i in range(int(samples - 1)):
        t1 = (i + 1) * step + t_start
        t2 = t1 - ds  # Second point BEFORE first point (Good for last point)
        x1 = f1(t1)
        x2 = f1(t2)
        y1 = f2(t1)
        y2 = f2(t2)

        # numerical derivatives
        dx1 = (x1 - x2) / ds
        dy1 = (y1 - y2) / ds

        # create curve
        a.append(
            [
                "C",
                [
                    coordx(x0 + (dx0 * third)),
                    coordy(y0 + (dy0 * third)),
                    coordx(x1 - (dx1 * third)),
                    coordy(y1 - (dy1 * third)),
                    coordx(x1),
                    coordy(y1),
                ],
            ]
        )
        t0 = t1  # Next segment's start is this segments end
        x0 = x1
        y0 = y1
        dx0 = dx1  # Assume the functions are smooth everywhere, so carry over the derivatives too
        dy0 = dy1
    return a


class ParamCurves(inkex.EffectExtension):
    def add_arguments(self, pars):
        pars.add_argument("--t_start", type=float, default=0.0, help="Start t-value")
        pars.add_argument("--t_end", type=float, default=1.0, help="End t-value")
        pars.add_argument(
            "--times2pi",
            type=inkex.Boolean,
            default=True,
            help="Multiply t-range by 2*pi",
        )
        pars.add_argument("--xleft", type=float, default=-1.0, help="x-value of left")
        pars.add_argument("--xright", type=float, default=1.0, help="x-value of right")
        pars.add_argument(
            "--ybottom", type=float, default=-1.0, help="y-value of bottom"
        )
        pars.add_argument("--ytop", type=float, default=1.0, help="y-value of top")
        pars.add_argument("-s", "--samples", type=int, default=30, help="Samples")
        pars.add_argument("--fofx", default="cos(3*t)", help="fx(t) for plotting")
        pars.add_argument("--fofy", default="sin(5*t)", help="fy(t) for plotting")
        pars.add_argument(
            "--remove", type=inkex.Boolean, default=True, help="Remove rectangle"
        )
        pars.add_argument(
            "--isoscale", type=inkex.Boolean, default=False, help="Isotropic scaling"
        )
        pars.add_argument("--drawaxis", type=inkex.Boolean, default=False)
        pars.add_argument("--tab", default="sampling")

    def effect(self):
        for node in self.svg.selection:
            if isinstance(node, inkex.Rectangle):
                # create new path with basic dimensions of selected rectangle
                newpath = inkex.PathElement()
                x = float(node.get("x"))
                y = float(node.get("y"))
                width = float(node.get("width"))
                height = float(node.get("height"))

                # copy attributes of rect
                newpath.style = node.style
                newpath.transform = node.transform

                # top and bottom were exchanged
                newpath.path = drawfunction(
                    self.options.t_start,
                    self.options.t_end,
                    self.options.xleft,
                    self.options.xright,
                    self.options.ybottom,
                    self.options.ytop,
                    self.options.samples,
                    width,
                    height,
                    x,
                    y + height,
                    self.options.fofx,
                    self.options.fofy,
                    self.options.times2pi,
                    self.options.isoscale,
                    self.options.drawaxis,
                )
                newpath.set("title", self.options.fofx + " " + self.options.fofy)

                # newpath.set('desc', '!func;' + self.options.fofx + ';' + self.options.fofy + ';'
                #                                      + `self.options.t_start` + ';'
                #                                      + `self.options.t_end` + ';'
                #                                      + `self.options.samples`)

                # add path into SVG structure
                node.getparent().append(newpath)
                # option whether to remove the rectangle or not.
                if self.options.remove:
                    node.getparent().remove(node)


if __name__ == "__main__":
    ParamCurves().run()



"""
VNPathElement, pathGeometry
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional

import inkex

from .base import VNBaseElement
from .styles import VNColor, VNGradient, pathStrokeStyle


@dataclass
class VNPathElement(VNBaseElement):
    """Path Element properties."""

    mask: int  # 0 or 1
    fillColor: Optional[VNColor]
    fillGradient: Optional[VNGradient]
    strokeStyle: Optional[pathStrokeStyle]
    # It's list because compoundPath has multiple pathGeometries
    pathGeometries: List[pathGeometry]


class pathGeometry:
    """path format in Linearity Curve(nodes)."""

    def __init__(self, closed: bool, nodes: List[Dict]):
        self.corner_radius: List[float] = []
        self.path = self.parse_nodes(closed, nodes)

    def parse_nodes(self, closed: bool, nodes: List[Dict]) -> inkex.Path:
        """Converts single pathGeometry data to inkex path."""
        path = None

        if closed:
            nodes.append(nodes[0])

        for node in nodes:
            # Path data is stored as  list of [inPoint, anchor, outPoint]
            # (plus some extra attributes for which we don't have enough data atm)
            anchor = inkex.Vector2d(node["anchorPoint"])
            if path is None:
                path = inkex.Path([inkex.paths.Move(*anchor)])
            else:
                inpt = inkex.Vector2d(node["inPoint"])
                anchor = inkex.Vector2d(node["anchorPoint"])

                # added * to add support for Inkscape 1.3
                if prev.is_close(outpt) and inpt.is_close(anchor):
                    path.append(inkex.paths.Line(*anchor))
                else:
                    path.append(inkex.paths.Curve(*outpt, *inpt, *anchor))

            prev = anchor
            outpt = inkex.Vector2d(node["outPoint"])

            # add corner radius to the list
            self.corner_radius.append(node["cornerRadius"])

        if closed:
            path.append(inkex.paths.ZoneClose())

        return path



#!/usr/bin/env python3
# coding=utf-8

# Written by Jabiertxof
# V.06

import os

from inkex import load_svg, TemplateExtension, Transform


class SeamlessPattern(TemplateExtension):
    """Generate a seamless pattern template"""

    multi_inx = True

    @classmethod
    def get_template(cls, **kwargs):
        name = "seamless_pattern.svg"
        path = os.path.dirname(os.path.realpath(__file__))
        return load_svg(os.path.join(path, name))

    def set_namedview(self, width, height, unit):
        width = self.options.width
        height = self.options.height
        factor = width / height
        clip_rect = self.svg.getElementById("clipPathRect")
        clip_rect.set("width", str(width))
        clip_rect.set("height", str(height))

        scale = Transform(scale=(width / 100, height / 100))
        self.svg.getElementById("designTop").transform = scale
        self.svg.getElementById("designBottom").transform = scale

        scale = (1, factor) if factor <= 1 else (1 / factor, 1)
        for child in self.svg.getElementById("designTop"):
            child.transform = Transform(scale=scale)

        text_preview = self.svg.getElementById("textPreview")
        if text_preview is not None:
            x = width / 100.0 / factor
            y = height / 100.0
            if factor <= 1:
                x *= factor
                y *= factor
            text_preview.transform = Transform(
                translate=(int(width) * 2, 0), scale=(x, y)
            )

        info_group = self.svg.getElementById("infoGroup")
        if info_group is not None:
            scale = 100 if factor <= 1 else 1000
            info_group.transform = Transform(
                scale=(width / scale, height / scale * factor)
            )

        sides = [(x, y) for y in (-height, 0, height) for x in (-width, 0, width)]
        for i, (x, y) in enumerate(sides):
            top = self.svg.getElementById("top{i}".format(i=i + 1))
            bottom = self.svg.getElementById("bottom{i}".format(i=i + 1))
            if top is not None and bottom is not None:
                bottom.transform = top.transform = Transform(translate=(x, y))

        clones = [
            (x, y) for x in (0, width, width * 2) for y in (0, height, height * 2)
        ]
        for i, (x, y) in enumerate(clones):
            preview = self.svg.getElementById("clonePreview{i}".format(i=i))
            if preview is not None:
                preview.transform = Transform(translate=(x, y))

        pattern_generator = self.svg.getElementById("fullPatternClone")
        if pattern_generator is not None:
            pattern_generator.transform = Transform(translate=(width * 2, -height))
            pattern_generator.set("inkscape:tile-cx", width / 2)
            pattern_generator.set("inkscape:tile-cy", height / 2)
            pattern_generator.set("inkscape:tile-w", width)
            pattern_generator.set("inkscape:tile-h", height)
            pattern_generator.set("inkscape:tile-x0", width)
            pattern_generator.set("inkscape:tile-y0", height)
            pattern_generator.set("width", width)
            pattern_generator.set("height", height)

        namedview = self.svg.namedview
        namedview.set("inkscape:document-units", "px")
        namedview.set("inkscape:cx", (width * 5.5) / 2)
        namedview.set("inkscape:cy", "0")
        namedview.set("inkscape:zoom", 1 / (width / 100))


if __name__ == "__main__":
    SeamlessPattern().run()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Manpreet Singh <manpreet.singh.dev@proton.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later


import math

import inkex


def process(svg: inkex.SvgDocumentElement) -> None:
    for elem in svg:
        if isinstance(elem, inkex.Defs):
            process(elem)
        if not isinstance(elem, inkex.ShapeElement):
            continue
        if isinstance(elem, inkex.elements._groups.GroupBase):
            enhance_group(elem)
            continue
        enhance_shape(elem)


def enhance_group(svg: inkex.elements._groups.GroupBase) -> None:
    group_elements = [e for e in svg]
    keep_group = (
        svg.getparent() is None
        or not isinstance(svg, inkex.Group)
        or any(s not in ("opacity") for s in svg.style)
        or any(a not in ("id", "transform", "style", "clip-path") for a in svg.keys())
        or (
            svg.get("clip-path") is not None
            and any(
                e.get("clip-path") is not None
                for e in group_elements
                if not isinstance(e, inkex.Group)
            )
        )
    )

    for elem in group_elements:
        if not isinstance(elem, inkex.ShapeElement):
            continue

        if not keep_group:
            elem.transform = svg.transform @ elem.transform
            elem.style["opacity"] = float(elem.style("opacity", default=1.0)) * float(
                svg.style.get("opacity", default=1.0)
            )
            if svg.clip is not None and not isinstance(
                elem, inkex.elements._groups.GroupBase
            ):
                elem.clip = svg.clip

        if isinstance(elem, inkex.elements._groups.GroupBase):
            enhance_group(elem)
            continue

        enhance_shape(elem)

    # Remove excessive groups
    if not keep_group:
        for elem in svg:
            svg.remove(elem)
            svg.addprevious(elem)
        svg.getparent().remove(svg)


def enhance_shape(svg: inkex.ShapeElement) -> None:
    # Enhance transforms
    tfm: inkex.Transform = svg.transform
    enhance_tfm(svg)
    baked_tfm = tfm @ -svg.transform
    enhance_gradients(svg, baked_tfm)
    enhance_clip(svg)


def enhance_clip(svg: inkex.ShapeElement) -> None:
    if svg.clip is None:
        return
    assert len(list(svg.clip)) == 1
    if isinstance(list(svg.clip)[0], inkex.Use):
        list(svg.clip)[0].transform = -svg.transform


def enhance_gradients(svg: inkex.ShapeElement, tfm: inkex.Transform) -> None:
    root: inkex.SvgDocumentElement = svg.getroottree().getroot()
    fill = root.getElementById(svg.style.get("fill", None))
    stroke = root.getElementById(svg.style.get("stroke", None))

    for grad in (fill, stroke):
        if not isinstance(grad, inkex.Gradient):
            continue
        gtfm = inkex.Transform(grad.get("gradientTransform"))
        grad.set("gradientTransform", tfm @ gtfm)


def enhance_tfm(svg: inkex.ShapeElement) -> None:
    if isinstance(svg, (inkex.TextElement, inkex.Image, inkex.Use)):
        return

    tfm: inkex.Transform = svg.transform
    svg.transform = inkex.Transform()

    if isinstance(svg, inkex.PathElement):
        if svg.get("inkscape:original-d", None):
            svg.original_path = svg.original_path.transform(tfm, inplace=True)
        # On stars / arcs, we preserve the transforms so that the
        # control points (inkscape:c1 etc.) are also transformed
        if svg.get("sodipodi:type", None) not in ("star", "arc"):
            svg.path = svg.path.transform(tfm, inplace=True)
        else:
            svg.transform = tfm
        return

    if isinstance(svg, inkex.Polygon):
        svg.path = svg.path.transform(tfm, inplace=True)
        return

    # Scaling is handled in the conversion step; this adjusts for uneven stroke width caused by shearing.
    if not (tfm.a or tfm.b) or not (tfm.c or tfm.d):
        svg.transform = tfm
        return

    scale = inkex.Transform(
        (math.hypot(tfm.a, tfm.b), 0, 0, math.hypot(tfm.c, tfm.d), 0, 0)
    )
    svg.transform = inkex.Transform(
        (
            tfm.a / scale.a,
            tfm.b / scale.a,
            tfm.c / scale.d,
            tfm.d / scale.d,
            tfm.e,
            tfm.f,
        )
    )

    if isinstance(svg, (inkex.Rectangle)):
        pos0 = scale.apply_to_point((svg.left, svg.top))
        pos1 = scale.apply_to_point((svg.left + svg.width, svg.top + svg.height))
        svg.set("x", min(pos1.x, pos0.x))
        svg.set("y", min(pos1.y, pos0.y))
        svg.set("width", abs(pos1.x - pos0.x))
        svg.set("height", abs(pos1.y - pos0.y))
        return

    if isinstance(svg, inkex.Ellipse):
        pos0 = scale.apply_to_point((svg.center - svg.radius))
        pos1 = scale.apply_to_point((svg.center + svg.radius))
        svg.radius = (pos1 - pos0) / 2
        svg.center = pos0 + svg.radius
        return

    raise NotImplementedError(f"Enhance Transform not implemented for {type(svg)}")



#
# Copyright 2021 Martin Owens <doctormo@gmail.com>
# Copyright 2022 Simon Duerr <dev@simonduerr.eu>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
Base module for all import web search source modules.
"""

import re
import os
import sys
import logging
import requests
import importlib

from cachecontrol import CacheControl, CacheControlAdapter
from cachecontrol.caches.file_cache import FileCache
from cachecontrol.heuristics import ExpiresAfter

from inkex.command import CommandNotFound, ProgramRunError, call
from collections import defaultdict

LICENSE_ICONS = os.path.join(os.path.dirname(__file__), "licenses")

LICENSES = {
    "cc-0": {
        "name": "CC0",
        "modules": ["nocopyright"],
        "url": "https://creativecommons.org/publicdomain/zero/1.0/",
        "overlay": "cc0.svg",
    },
    "cc-by-3.0": {
        "name": "CC-BY 3.0 Unported",
        "modules": ["by"],
        "url": "https://creativecommons.org/licenses/by/3.0/",
        "overlay": "cc-by.svg",
    },
    "cc-by-4.0": {
        "name": "CC-BY 4.0 Unported",
        "modules": ["by"],
        "url": "https://creativecommons.org/licenses/by/4.0/",
        "overlay": "cc-by.svg",
    },
    "cc-by-sa-4.0": {
        "name": "CC-BY SA 4.0",
        "modules": ["by", "sa"],
        "url": "https://creativecommons.org/licenses/by-sa/4.0/",
        "overlay": "cc-by-sa.svg",
    },
    "cc-by-sa-3.0": {
        "name": "CC-BY SA 3.0",
        "modules": ["by", "sa"],
        "url": "https://creativecommons.org/licenses/by-sa/3.0/",
        "overlay": "cc-by-sa.svg",
    },
    "cc-by-nc-sa-4.0": {
        "name": "CC-BY NC SA 4.0",
        "modules": ["by", "sa", "nc"],
        "url": "https://creativecommons.org/licenses/by-nc-sa/4.0/",
        "overlay": "cc-by-nc-sa.svg",
    },
    "cc-by-nc-sa-3.0": {
        "name": "CC-BY NC SA 3.0",
        "modules": ["by", "sa", "nc"],
        "url": "https://creativecommons.org/licenses/by-nc-sa/3.0/",
        "overlay": "cc-by-nc-sa.svg",
    },
    "cc-by-nc-3.0": {
        "name": "CC-BY NC 3.0",
        "modules": ["by", "nc"],
        "url": "https://creativecommons.org/licenses/by-nc/3.0/",
        "overlay": "cc-by-nc.svg",
    },
    "cc-by-nd-3.0": {
        "name": "CC-BY ND 3.0",
        "modules": ["by", "nd"],
        "url": "https://creativecommons.org/licenses/by-nd/3.0/",
        "overlay": "cc-by-nd.svg",
    },
    "gpl-2": {
        "name": "GPLv2",
        "modules": ["retaincopyrightnotice", "sa"],
        "url": "https://www.gnu.org/licenses/old-licenses/gpl-2.0.txt",
        "overlay": "gpl.svg",
    },
    "gpl-3": {
        "name": "GPLv3",
        "modules": ["retaincopyrightnotice", "sa"],
        "url": "https://www.gnu.org/licenses/gpl-3.0.txt",
        "overlay": "gpl.svg",
    },
    "agpl-3": {
        "name": "AGPLv3",
        "modules": ["retaincopyrightnotice", "sa"],
        "url": "https://www.gnu.org/licenses/agpl-3.0.txt",
        "overlay": "gpl.svg",
    },
    "mit": {
        "name": "MIT",
        "modules": ["retaincopyrightnotice"],
        "url": "https://mit-license.org/",
        "overlay": "mit.svg",
    },
    "asl": {
        "name": "Apache License",
        "modules": ["retaincopyrightnotice"],
        "url": "https://www.apache.org/licenses/LICENSE-2.0.txt",
        "overlay": "asl.svg",
    },
    "bsd": {
        "name": "BSD",
        "modules": ["retaincopyrightnotice", "noendorsement"],
        "url": "https://opensource.org/licenses/BSD-3-Clause",
        "overlay": "bsd.svg",
    },
}


class RemotePage:
    """Lazy access to paging systems"""

    icon = "sources/next_page.svg"
    string = property(lambda self: "Next Page")

    def __init__(self, remote, func):
        self.func = func
        self.remote = remote

    def get_next_page(self):
        for info in self.func():
            yield self.remote.result_to_cls(info)


class RemoteFile:
    """Lazy access to remote files"""

    icon = property(lambda self: self.remote.to_local_file(self.info["thumbnail"]))
    get_file = lambda self: self.remote.to_local_file(self.info["file"])

    def __init__(self, remote, info):
        for field in ("name", "thumbnail", "license", "file"):
            if field not in info:
                raise ValueError(f"Field {field} not provided in RemoteFile package")
        self.info = info
        self.remote = remote

    @property
    def string(self):
        return self.info["name"]

    @property
    def license(self):
        return self.info["license"]

    def get_overlay(self):
        return self.license_info["overlay"]

    @property
    def license_info(self):
        return LICENSES.get(
            self.license,
            {
                "name": "Unknown",
                "url": self.info.get("descriptionurl", ""),
                "modules": [],
                "overlay": "unknown.svg",
            },
        )

    @property
    def author(self):
        return self.info["author"]


class RemoteSource:
    """A remote source of svg images which can be searched and downloaded"""

    # These are the properties that should be overridden in your class
    name = None
    icon = None
    file_cls = RemoteFile
    page_cls = RemotePage
    is_default = False
    is_enabled = True

    @classmethod
    def load(cls, name):
        """Load the file or directory of remote sources"""
        if os.path.isfile(name):
            sys.path, sys_path = [os.path.dirname(name)] + sys.path, sys.path
            try:
                importlib.import_module(os.path.basename(name).rsplit(".", 1)[0])
            except ImportError:
                logging.error(f"Failed to load module: {name}")
            sys.path = sys_path
        elif os.path.isdir(name):
            for child in os.listdir(name):
                if not child.startswith("_") and child.endswith(".py"):
                    cls.load(os.path.join(name, child))

    def search(self, query, tags=[]):
        """
        Search for the given query and yield basic informational blocks t hand to file_cls.

        Required fields per yielded object are: name, license, thumbnail and file.
        Optional fields are: id, summary, author, created, popularity
        """
        raise NotImplementedError(
            "You must implement a search function for this remote source!"
        )

    sources = {}

    def __init_subclass__(cls):
        if cls != RemoteSource:
            cls.sources[cls.__name__] = cls

    def __init__(self, cache_dir):
        self.session = requests.session()
        self.cache_dir = cache_dir
        try:
            self.session.mount(
                "https://",
                CacheControlAdapter(
                    cache=FileCache(cache_dir),
                    heuristic=ExpiresAfter(days=5),
                ),
            )
        except ImportError:
            # This happens when python-lockfile is missing, disable cachecontrol
            pass

    def __del__(self):
        self.session.close()

    def file_search(self, query):
        """Search for extension packages"""
        for info in self.search(query):
            yield self.result_to_cls(info)

    def result_to_cls(self, info):
        if callable(info):
            return self.page_cls(self, info)
        return self.file_cls(self, info)

    def to_local_file(self, url):
        """Get a remote url and turn it into a local file"""
        filepath = os.path.join(self.cache_dir, url.split("/")[-1])
        headers = {"User-Agent": "Inkscape"}
        try:
            remote = self.session.get(
                url, headers=headers
            )  # needs UserAgent otherwise many 403 or 429 for wiki commons
        except requests.exceptions.RequestException as err:
            return None
        except ConnectionError as err:
            return None
        except requests.exceptions.RequestsWarning:
            pass

        if remote and remote.status_code == 200:
            with open(filepath, "wb") as fhl:
                # If we don't have data, return None (instead of empty file)
                if fhl.write(remote.content):
                    return filepath
        return None



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2012 Jabiertxo Arraiza, jabier.arraiza@marker.es
# Copyright (C) 2016 su_v, <suv-sf@users.sf.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Version 0.6 - DPI Switcher

This extension scales a document to fit different SVG DPI -90/96-

Changes since v0.5:
    - transform all top-level containers and graphics elements
    - support scientific notation in SVG lengths
    - fix scaling with existing matrix()
    - support different units for document width, height attributes
    - improve viewBox support (syntax, offset)
    - support common cases of text-put-on-path in SVG root
    - support common cases of <use> references in SVG root
    - examples from http://tavmjong.free.fr/INKSCAPE/UNITS/ tested

TODO:
    - check grids/guides created with 0.91:
      http://tavmjong.free.fr/INKSCAPE/UNITS/units_mm_nv_90dpi.svg
    - check <symbol> instances
    - check more <use> and text-on-path cases (reverse scaling needed?)
    - scale perspective of 3dboxes

"""

import re
import math
import inkex
from inkex import Use, TextElement, transforms

# globals
SKIP_CONTAINERS = [
    "defs",
    "glyph",
    "marker",
    "mask",
    "missing-glyph",
    "pattern",
    "symbol",
]
CONTAINER_ELEMENTS = [
    "a",
    "g",
    "switch",
]
GRAPHICS_ELEMENTS = [
    "circle",
    "ellipse",
    "image",
    "line",
    "path",
    "polygon",
    "polyline",
    "rect",
    "text",
    "use",
]


def is_3dbox(element):
    """Check whether element is an Inkscape 3dbox type."""
    return element.get("sodipodi:type") == "inkscape:box3d"


def is_text_on_path(element):
    """Check whether text element is put on a path."""
    if isinstance(element, TextElement):
        text_path = element.find("svg:textPath")
        if text_path is not None and len(text_path):
            return True
    return False


def is_sibling(element1, element2):
    """Check whether element1 and element2 are siblings of same parent."""
    return element2 in element1.getparent()


def is_in_defs(doc, element):
    """Check whether element is in defs."""
    if element is not None:
        defs = doc.find("defs")
        if defs is not None:
            return element in defs.iterdescendants()
    return False


def check_3dbox(svg, element, scale_x, scale_y):
    """Check transformation for 3dbox element."""
    skip = False
    if skip:
        # 3dbox elements ignore preserved transforms
        # FIXME: manually update geometry of 3dbox?
        pass
    return skip


def check_text_on_path(svg, element, scale_x, scale_y):
    """Check whether to skip scaling a text put on a path."""
    skip = False
    path = element.find("textPath").href
    if not is_in_defs(svg, path):
        if is_sibling(element, path):
            # skip common element scaling if both text and path are siblings
            skip = True
            # scale offset
            if "transform" in element.attrib:
                element.transform.add_scale(scale_x, scale_y)
            # scale font size
            mat = inkex.Transform("scale({},{})".format(scale_x, scale_y)).matrix
            det = abs(mat[0][0] * mat[1][1] - mat[0][1] * mat[1][0])
            descrim = math.sqrt(abs(det))
            prop = "font-size"
            # outer text
            sdict = dict(inkex.Style.parse_str(element.get("style")))
            if prop in sdict:
                sdict[prop] = float(sdict[prop]) * descrim
                element.set("style", str(inkex.Style(sdict)))
            # inner tspans
            for child in element.iterdescendants():
                if isinstance(element, inkex.Tspan):
                    sdict = dict(inkex.Style.parse_str(child.get("style")))
                    if prop in sdict:
                        sdict[prop] = float(sdict[prop]) * descrim
                        child.set("style", str(inkex.Style(sdict)))
    return skip


def check_use(svg, element, scale_x, scale_y):
    """Check whether to skip scaling an instantiated element (<use>)."""
    skip = False
    path = element.href
    if not is_in_defs(svg, path):
        if is_sibling(element, path):
            skip = True
            # scale offset
            if "transform" in element.attrib:
                element.transform.add_scale(scale_x, scale_y)
    return skip


class DPISwitcher(inkex.EffectExtension):
    multi_inx = True
    factor_a = 90.0 / 96.0
    factor_b = 96.0 / 90.0
    units = "px"

    def add_arguments(self, pars):
        pars.add_argument(
            "--switcher", type=str, default="0", help="Select the DPI switch you want"
        )

    # dictionaries of unit to user unit conversion factors
    __uuconvLegacy = {
        "in": 90.0,
        "pt": 1.25,
        "px": 1.0,
        "mm": 3.5433070866,
        "cm": 35.433070866,
        "m": 3543.3070866,
        "km": 3543307.0866,
        "pc": 15.0,
        "yd": 3240.0,
        "ft": 1080.0,
    }
    __uuconv = {
        "in": 96.0,
        "pt": 1.33333333333,
        "px": 1.0,
        "mm": 3.77952755913,
        "cm": 37.7952755913,
        "m": 3779.52755913,
        "km": 3779527.55913,
        "pc": 16.0,
        "yd": 3456.0,
        "ft": 1152.0,
    }

    def parse_length(self, length, percent=False):
        """Parse SVG length."""
        if self.options.switcher == "0":  # dpi90to96
            known_units = list(self.__uuconvLegacy)
        else:  # dpi96to90
            known_units = list(self.__uuconv)
        if percent:
            unitmatch = re.compile("(%s)$" % "|".join(known_units + ["%"]))
        else:
            unitmatch = re.compile("(%s)$" % "|".join(known_units))
        param = re.compile(
            r"(([-+]?[0-9]+(\.[0-9]*)?|[-+]?\.[0-9]+)([eE][-+]?[0-9]+)?)"
        )
        p = param.match(length)
        u = unitmatch.search(length)
        val = 100  # fallback: assume default length of 100
        unit = "px"  # fallback: assume 'px' unit
        if p:
            val = float(p.string[p.start() : p.end()])
        if u:
            unit = u.string[u.start() : u.end()]
        return val, unit

    def convert_length(self, val, unit):
        """Convert length to self.units if unit differs."""
        doc_unit = self.units or "px"
        if unit != doc_unit:
            if self.options.switcher == "0":  # dpi90to96
                val_px = val * self.__uuconvLegacy[unit]
                val = val_px / (
                    self.__uuconvLegacy[doc_unit] / self.__uuconvLegacy["px"]
                )
                unit = doc_unit
            else:  # dpi96to90
                val_px = val * self.__uuconv[unit]
                val = val_px / (self.__uuconv[doc_unit] / self.__uuconv["px"])
                unit = doc_unit
        return val, unit

    def check_attr_unit(self, element, attr, unit_list):
        """Check unit of attribute value, match to units in *unit_list*."""
        if attr in element.attrib:
            unit = self.parse_length(element.get(attr), percent=True)[1]
            return unit in unit_list

    def scale_attr_val(self, element, attr, unit_list, factor):
        """Scale attribute value if unit matches one in *unit_list*."""
        if attr in element.attrib:
            val, unit = self.parse_length(element.get(attr), percent=True)
            if unit in unit_list:
                element.set(attr, "{}{}".format(val * factor, unit))

    def scale_root(self, unit_exponent=1.0):
        """Scale all top-level elements in SVG root."""

        # update viewport
        width_num = self.parse_length(self.svg.get("width"))[0]
        height_num = self.convert_length(*self.parse_length(self.svg.get("height")))[0]
        width_doc = width_num * self.factor_a * unit_exponent
        height_doc = height_num * self.factor_a * unit_exponent

        svg = self.svg
        if svg.get("height"):
            svg.set("height", str(height_doc))
        if svg.get("width"):
            svg.set("width", str(width_doc))

        # update viewBox
        if svg.get("viewBox"):
            viewboxstring = re.sub(" +|, +|,", " ", svg.get("viewBox"))
            viewboxlist = [float(i) for i in viewboxstring.strip().split(" ", 4)]
            svg.set(
                "viewBox",
                "{} {} {} {}".format(*[(val * self.factor_a) for val in viewboxlist]),
            )

        # update guides, grids
        if self.options.switcher == "1":
            # FIXME: dpi96to90 only?
            self.scale_guides()
            self.scale_grid()

        for element in svg:  # iterate all top-level elements of SVGRoot
            # init variables
            tag = element.TAG
            width_scale = self.factor_a
            height_scale = self.factor_a

            if tag in GRAPHICS_ELEMENTS or tag in CONTAINER_ELEMENTS:
                # test for specific elements to skip from scaling
                if is_3dbox(element):
                    if check_3dbox(svg, element, width_scale, height_scale):
                        continue
                if is_text_on_path(element):
                    if check_text_on_path(svg, element, width_scale, height_scale):
                        continue
                if isinstance(element, Use):
                    if check_use(svg, element, width_scale, height_scale):
                        continue

                # relative units ('%') in presentation attributes
                for attr in ["width", "height"]:
                    self.scale_attr_val(element, attr, ["%"], 1.0 / self.factor_a)
                for attr in ["x", "y"]:
                    self.scale_attr_val(element, attr, ["%"], 1.0 / self.factor_a)

                # set preserved transforms on top-level elements
                if width_scale != 1.0 and height_scale != 1.0:
                    scale_mat = transforms.Transform(scale=(width_scale, height_scale))
                    element.transform = scale_mat @ element.transform

    def scale_element(self, elem):
        pass  # TODO: optionally scale graphics elements only?

    def scale_guides(self):
        """Scale the guidelines"""
        for guide in self.svg.namedview.get_guides():
            point = guide.get("position").split(",")
            guide.set(
                "position",
                str(float(point[0].strip()) * self.factor_a)
                + ","
                + str(float(point[1].strip()) * self.factor_a),
            )

    def scale_grid(self):
        """Scale the inkscape grid"""
        grids = self.svg.xpath("//inkscape:grid")
        for grid in grids:
            grid.set("units", "px")
            if grid.get("spacingx"):
                spacingx = (
                    str(
                        float(re.sub("[a-zA-Z]", "", grid.get("spacingx")))
                        * self.factor_a
                    )
                    + "px"
                )
                grid.set("spacingx", str(spacingx))
            if grid.get("spacingy"):
                spacingy = (
                    str(
                        float(re.sub("[a-zA-Z]", "", grid.get("spacingy")))
                        * self.factor_a
                    )
                    + "px"
                )
                grid.set("spacingy", str(spacingy))
            if grid.get("originx"):
                originx = (
                    str(
                        float(re.sub("[a-zA-Z]", "", grid.get("originx")))
                        * self.factor_a
                    )
                    + "px"
                )
                grid.set("originx", str(originx))
            if grid.get("originy"):
                originy = (
                    str(
                        float(re.sub("[a-zA-Z]", "", grid.get("originy")))
                        * self.factor_a
                    )
                    + "px"
                )
                grid.set("originy", str(originy))

    def effect(self):
        svg = self.svg
        if self.options.switcher == "0":
            self.factor_a = 96.0 / 90.0
            self.factor_b = 90.0 / 96.0
        svg.namedview.set("inkscape:document-units", "px")
        self.units = self.parse_length(svg.get("width"))[1]
        unit_exponent = 1.0
        if self.units and self.units != "px" and self.units != "" and self.units != "%":
            if self.options.switcher == "0":
                unit_exponent = 1.0 / (self.factor_a / self.__uuconv[self.units])
            else:
                unit_exponent = 1.0 / (self.factor_a / self.__uuconvLegacy[self.units])
        self.scale_root(unit_exponent)


if __name__ == "__main__":
    DPISwitcher().run()



#
# Copyright (C) 2019 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Utilities functions for inkscape extension manager.
"""

import os

from collections import defaultdict
from email.parser import FeedParser
from appdirs import user_cache_dir

from inkex.inx import InxFile
from inkex.command import inkscape, ProgramRunError, CommandNotFound

DEFAULT_VERSION = "1.1"
CACHE_DIR = user_cache_dir("inkscape-extension-manager", "Inkscape")
INKSCAPE_VERSION = os.environ.get("INKSCAPE_VERSION", None)
INKSCAPE_PROFILE = os.environ.get("INKSCAPE_PROFILE_DIR", None)

# This directory can be passed and used by inkscape to override it's
# profile directory for extensions
INKSCAPE_EXTENSIONS = os.environ.get("INKSCAPE_EXTENSIONS_DIR", None)

if not INKSCAPE_PROFILE and "VIRTUAL_ENV" in os.environ:
    INKSCAPE_PROFILE = os.path.dirname(os.environ["VIRTUAL_ENV"])

if not INKSCAPE_PROFILE and "APPDATA" in os.environ:
    INKSCAPE_PROFILE = os.path.join(os.environ["APPDATA"], "inkscape")

if not INKSCAPE_PROFILE and not INKSCAPE_EXTENSIONS:
    raise ImportError("The Inkscape profile directory isn't set!")

if not os.path.isdir(CACHE_DIR):
    os.makedirs(CACHE_DIR)

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
ICON_SEP = ("-" * 12) + "SVGICON" + ("-" * 12)


def _pythonpath():
    for pth in os.environ.get("PYTHONPATH", "").split(":"):
        if os.path.isdir(pth):
            yield pth


def get_user_directory():
    """Return the user directory where extensions are stored."""
    if INKSCAPE_EXTENSIONS:
        return os.path.abspath(INKSCAPE_EXTENSIONS)

    if "INKSCAPE_PROFILE_DIR" in os.environ:
        return os.path.abspath(
            os.path.expanduser(
                os.path.join(os.environ["INKSCAPE_PROFILE_DIR"], "extensions")
            )
        )

    home = os.path.expanduser("~")
    for pth in _pythonpath():
        if pth.startswith(home):
            return pth


def get_inkscape_directory():
    """Return the system directory where inkscape's core is."""
    for pth in _pythonpath():
        if os.path.isdir(os.path.join(pth, "inkex")):
            return pth


def get_inkscape_version():
    """Attempt to detect the inkscape version"""
    try:
        line = inkscape(version=True, svg_file=None)
    except (ProgramRunError, CommandNotFound):
        return DEFAULT_VERSION
    if isinstance(line, bytes):
        line = line.decode("utf8")
    (major, minor) = line.strip().split(" ")[1].split(".")[:2]
    return "{}.{}".format(int(major), int(minor.split("-")[0]))


def format_requires(string):
    """Get a version requires."""
    primary = string.split("; ", 1)[0]
    if "(" in primary:
        primary, version = primary.split("(", 1)
        return (primary.strip(), version.strip(") "))
    return (primary.strip(), None)


def parse_metadata(data):
    """
    Convert older email based meta data into a newer json format,
    See PEP 566 for details.
    """
    feed_parser = FeedParser()
    feed_parser.feed(data)
    metadata = feed_parser.close()

    def getdict():
        """Multi-dimentional dictionary"""
        return defaultdict(getdict)

    ret = defaultdict(getdict)

    ret["description"] = metadata.get_payload()
    ret["requires"] = [
        format_requires(m) for m in metadata.get_all("Requires-Dist", [])
    ]

    for key, value in metadata.items():
        if key == "Home-page":
            ret["extensions"]["python.details"]["project_urls"]["Home"] = value
        elif key == "Classifier":
            ret["classifiers"] = list(ret["classifiers"])
            ret["classifiers"].append(value)
        else:
            ret[key.lower().replace("-", "_")] = value

    return ret


def clean_author(data):
    """Clean the author so it has consistant keys"""
    for contact in (
        data.get("extensions", {})
        .get("python.details", {})
        .get("contacts", [])
    ):
        if contact["role"] == "author":
            data["author"] = contact["name"]
            data["author_email"] = contact.get("email", "")
            if "<" in data["author"]:
                data["author"], other = data["author"].split("<", 1)
                if "@" in other and not data["author_email"]:
                    data["author_email"] = data["author"].split(">")[0]
    return data


class ExtensionInx(InxFile):
    """Information about an extension specifically"""

    ident = property(lambda self: super().ident or f"[no-id] {self.filename}")
    name = property(lambda self: super().name or f"[unnamed] {self.ident}")

    @property
    def menu(self):
        menu = self.xml.find_one("effect/effects-menu")
        if menu is not None and menu.get("hidden", "false") == "true":
            return ["_hidden", self.name]
        return super().menu



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2010 Alvin Penner, penner@vaxxine.com
#
# - Voronoi Diagram algorithm and C code by Steven Fortune, 1987, http://ect.bell-labs.com/who/sjf/
# - Python translation to file voronoi.py by Bill Simons, 2005, http://www.oxfish.com/
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import random

import inkex
from inkex import PathElement, Pattern
from inkex.localization import inkex_gettext as _

import voronoi


def clip_line(x1, y1, x2, y2, w, h):
    if x1 < 0 and x2 < 0:
        return [0, 0, 0, 0]
    if x1 > w and x2 > w:
        return [0, 0, 0, 0]
    if x1 < 0:
        y1 = (y1 * x2 - y2 * x1) / (x2 - x1)
        x1 = 0
    if x2 < 0:
        y2 = (y1 * x2 - y2 * x1) / (x2 - x1)
        x2 = 0
    if x1 > w:
        y1 = y1 + (w - x1) * (y2 - y1) / (x2 - x1)
        x1 = w
    if x2 > w:
        y2 = y1 + (w - x1) * (y2 - y1) / (x2 - x1)
        x2 = w
    if y1 < 0 and y2 < 0:
        return [0, 0, 0, 0]
    if y1 > h and y2 > h:
        return [0, 0, 0, 0]
    if x1 == x2 and y1 == y2:
        return [0, 0, 0, 0]
    if y1 < 0:
        x1 = (x1 * y2 - x2 * y1) / (y2 - y1)
        y1 = 0
    if y2 < 0:
        x2 = (x1 * y2 - x2 * y1) / (y2 - y1)
        y2 = 0
    if y1 > h:
        x1 = x1 + (h - y1) * (x2 - x1) / (y2 - y1)
        y1 = h
    if y2 > h:
        x2 = x1 + (h - y1) * (x2 - x1) / (y2 - y1)
        y2 = h
    return [x1, y1, x2, y2]


class VoronoiFill(inkex.EffectExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument(
            "--size", type=int, default=10, help="Average size of cell (px)"
        )
        pars.add_argument("--border", type=int, default=0, help="Size of Border (px)")

    def effect(self):
        if not self.options.ids:
            return inkex.errormsg(_("Please select an object"))
        scale = self.svg.unittouu("1px")  # convert to document units
        self.options.size *= scale
        self.options.border *= scale
        obj = self.svg.selection.first()
        bbox = obj.bounding_box()
        mat = obj.composed_transform().matrix
        pattern = self.svg.defs.add(Pattern())
        pattern.set_random_id("Voronoi")
        pattern.set("width", str(bbox.width))
        pattern.set("height", str(bbox.height))
        pattern.set("patternUnits", "userSpaceOnUse")
        pattern.patternTransform.add_translate(
            bbox.left - mat[0][2], bbox.top - mat[1][2]
        )

        # generate random pattern of points
        c = voronoi.Context()
        pts = []
        b = float(self.options.border)  # width of border
        for i in range(
            int(bbox.width * bbox.height / self.options.size / self.options.size)
        ):
            x = random.random() * bbox.width
            y = random.random() * bbox.height
            if b > 0:  # duplicate border area
                pts.append(voronoi.Site(x, y))
                if x < b:
                    pts.append(voronoi.Site(x + bbox.width, y))
                    if y < b:
                        pts.append(voronoi.Site(x + bbox.width, y + bbox.height))
                    if y > bbox.height - b:
                        pts.append(voronoi.Site(x + bbox.width, y - bbox.height))
                if x > bbox.width - b:
                    pts.append(voronoi.Site(x - bbox.width, y))
                    if y < b:
                        pts.append(voronoi.Site(x - bbox.width, y + bbox.height))
                    if y > bbox.height - b:
                        pts.append(voronoi.Site(x - bbox.width, y - bbox.height))
                if y < b:
                    pts.append(voronoi.Site(x, y + bbox.height))
                if y > bbox.height - b:
                    pts.append(voronoi.Site(x, y - bbox.height))
            elif x > -b and y > -b and x < bbox.width + b and y < bbox.height + b:
                pts.append(voronoi.Site(x, y))  # leave border area blank
            # dot = pattern.add(inkex.Rectangle())
            # dot.set('x', str(x-1))
            # dot.set('y', str(y-1))
            # dot.set('width', '2')
            # dot.set('height', '2')
        if len(pts) < 3:
            return inkex.errormsg("Please choose a larger object, or smaller cell size")

        # plot Voronoi diagram
        sl = voronoi.SiteList(pts)
        voronoi.voronoi(sl, c)
        path = ""
        for edge in c.edges:
            if edge[1] >= 0 and edge[2] >= 0:  # two vertices
                [x1, y1, x2, y2] = clip_line(
                    c.vertices[edge[1]][0],
                    c.vertices[edge[1]][1],
                    c.vertices[edge[2]][0],
                    c.vertices[edge[2]][1],
                    bbox.width,
                    bbox.height,
                )
            elif edge[1] >= 0:  # only one vertex
                if c.lines[edge[0]][1] == 0:  # vertical line
                    xtemp = c.lines[edge[0]][2] / c.lines[edge[0]][0]
                    if c.vertices[edge[1]][1] > bbox.height / 2:
                        ytemp = bbox.height
                    else:
                        ytemp = 0
                else:
                    xtemp = bbox.width
                    ytemp = (
                        c.lines[edge[0]][2] - bbox.width * c.lines[edge[0]][0]
                    ) / c.lines[edge[0]][1]
                [x1, y1, x2, y2] = clip_line(
                    c.vertices[edge[1]][0],
                    c.vertices[edge[1]][1],
                    xtemp,
                    ytemp,
                    bbox.width,
                    bbox.height,
                )
            elif edge[2] >= 0:  # only one vertex
                if edge[0] >= len(c.lines):
                    xtemp = 0
                    ytemp = 0
                elif c.lines[edge[0]][1] == 0:  # vertical line
                    xtemp = c.lines[edge[0]][2] / c.lines[edge[0]][0]
                    if c.vertices[edge[2]][1] > bbox.height / 2:
                        ytemp = bbox.height
                    else:
                        ytemp = 0
                else:
                    xtemp = 0
                    ytemp = c.lines[edge[0]][2] / c.lines[edge[0]][1]
                [x1, y1, x2, y2] = clip_line(
                    xtemp,
                    ytemp,
                    c.vertices[edge[2]][0],
                    c.vertices[edge[2]][1],
                    bbox.width,
                    bbox.height,
                )
            if x1 or x2 or y1 or y2:
                path += "M %.3f,%.3f %.3f,%.3f " % (x1, y1, x2, y2)

        patternstyle = {"stroke": "#000000", "stroke-width": str(scale)}
        attribs = {"d": path, "style": str(inkex.Style(patternstyle))}
        pattern.append(PathElement(**attribs))

        # link selected object to pattern
        obj.style["fill"] = pattern
        if isinstance(obj, inkex.Group):
            for node in obj:
                node.style["fill"] = pattern


if __name__ == "__main__":
    VoronoiFill().run()



#!/usr/bin/env python3
"""Randomise the selected item's colours using hsl colorspace"""

from random import randrange, uniform, seed
import inkex
from inkex.localization import inkex_gettext as _


def _rand(
    limit, value, roof=255, method=randrange, circular=False, deterministic=False
):
    """Generates a random number which is less than limit % away from value, using the method
    supplied."""
    if deterministic:
        if isinstance(value, float):
            seed(int(value * 1000))
        else:
            seed(value)
    limit = roof * float(limit) / 100
    limit /= 2
    max_ = type(roof)(value + limit)
    min_ = type(roof)(value - limit)
    if not (circular):
        if max_ > roof:
            min_ -= max_ - roof
            max_ = roof
        if min_ < 0:
            max_ -= min_
            min_ = 0
        return method(min_, max_)
    return method(min_, max_) % roof


class Randomize(inkex.ColorExtension):
    """Randomize the colours of all objects"""

    deterministic_output = False

    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("-y", "--hue_range", type=int, default=0, help="Hue range")
        pars.add_argument(
            "-t", "--saturation_range", type=int, default=0, help="Saturation range"
        )
        pars.add_argument(
            "-m", "--lightness_range", type=int, default=0, help="Lightness range"
        )
        pars.add_argument(
            "-o", "--opacity_range", type=int, default=0, help="Opacity range"
        )

    def _rand(self, limit, value, roof=255, method=randrange, circular=False):
        return _rand(
            limit,
            value,
            roof,
            method,
            circular,
            deterministic=self.deterministic_output,
        )

    def modify_color(self, name, color):
        hsl = color.to_hsl()
        if self.options.hue_range > 0:
            hsl.hue = int(self._rand(self.options.hue_range, hsl.hue, circular=True))
        if self.options.saturation_range > 0:
            hsl.saturation = int(
                self._rand(self.options.saturation_range, hsl.saturation)
            )
        if self.options.lightness_range > 0:
            hsl.lightness = int(self._rand(self.options.lightness_range, hsl.lightness))
        return hsl.to_rgb()

    def modify_opacity(self, name, opacity):
        if name != "opacity":
            return opacity
        try:
            opacity = float(opacity)
        except ValueError:
            self.msg(_("Ignoring unusual opacity value: {}").format(opacity))
            return opacity
        orange = self.options.opacity_range
        if orange > 0:
            return self._rand(orange, opacity, roof=1.0, method=uniform)
        return opacity


if __name__ == "__main__":
    Randomize().run()



# SPDX-FileCopyrightText: 2023 Software Freedom Conservancy <info@sfconservancy.org>
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Commonly used functions."""

from inkaf.parser.extract import AFExtractor, NormalFATFile
from inkaf.parser.parse import AFParser
from inkaf.parser.types import Document
from lxml import etree
from io import BytesIO


def to_pretty_xml(xml_string: bytes) -> bytes:
    """Return a pretty xml string with newlines and indentation."""
    # from https://stackoverflow.com/a/3974112/1320237
    # and https://stackoverflow.com/a/9612463/1320237
    parser = etree.XMLParser(remove_blank_text=True)
    file = BytesIO()
    file.write(xml_string)
    file.seek(0)
    element = etree.parse(file, parser)
    return etree.tostring(element, pretty_print=True)


def extract(extractor: AFExtractor, doc: str = "doc.dat") -> bytes:
    assert extractor.content is not None, (
        f"Extractor content not initialized. '{doc}' extraction failed."
    )

    if doc not in extractor.content.files:
        revision = extractor.get_head_revision(doc)
        if revision is None:
            raise KeyError(doc)
        if not isinstance(revision, NormalFATFile):
            raise TypeError(f"Unexpected revision type: {type(revision)}")
        if revision.compressed_size == 0:
            raise RuntimeError(f"Empty revision: {revision}")

        file_content = extractor.extract(revision)

        assert extractor.content is not None
        extractor.content.files[doc] = file_content

    return extractor.content.files[doc]


def parse(extractor: AFExtractor, doc: str = "doc.dat") -> Document:
    file_content = extract(extractor, doc)
    result = AFParser(BytesIO(file_content))
    return result.parse()


class DuplicateFilter(object):
    def __init__(self):
        self.msgs = set()

    def filter(self, record):
        rv = record.msg not in self.msgs
        self.msgs.add(record.msg)
        return rv


__all__ = ["to_pretty_xml", "parse", "DuplicateFilter"]



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2007 Terry Brown, terry_n_brown@yahoo.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
from __future__ import absolute_import, unicode_literals

import inkex
from inkex import Use, Rectangle
from inkex.base import SvgOutputMixin
from inkex.localization import inkex_gettext as _


class Nup(inkex.OutputExtension, SvgOutputMixin):
    """N-up Layout generator"""

    def add_arguments(self, pars):
        pars.add_argument("--unit", default="px")
        pars.add_argument("--rows", type=int, default=2)
        pars.add_argument("--cols", type=int, default=2)
        pars.add_argument("--paddingTop", type=float)
        pars.add_argument("--paddingBottom", type=float)
        pars.add_argument("--paddingLeft", type=float)
        pars.add_argument("--paddingRight", type=float)
        pars.add_argument("--marginTop", type=float)
        pars.add_argument("--marginBottom", type=float)
        pars.add_argument("--marginLeft", type=float)
        pars.add_argument("--marginRight", type=float)
        pars.add_argument("--pgMarginTop", type=float)
        pars.add_argument("--pgMarginBottom", type=float)
        pars.add_argument("--pgMarginLeft", type=float)
        pars.add_argument("--pgMarginRight", type=float)
        pars.add_argument("--pgSizeX", type=float)
        pars.add_argument("--pgSizeY", type=float)
        pars.add_argument("--sizeX", type=float)
        pars.add_argument("--sizeY", type=float)
        pars.add_argument("--calculateSize", type=inkex.Boolean, default=True)
        pars.add_argument("--showHolder", type=inkex.Boolean, default=True)
        pars.add_argument("--showCrosses", type=inkex.Boolean, default=True)
        pars.add_argument("--showInner", type=inkex.Boolean, default=True)
        pars.add_argument("--showOuter", type=inkex.Boolean, default=False)
        pars.add_argument("--showInnerBox", type=inkex.Boolean, default=False)
        pars.add_argument("--showOuterBox", type=inkex.Boolean, default=False)
        pars.add_argument("--tab")

    def save(self, stream):
        show_list = []
        for i in [
            "showHolder",
            "showCrosses",
            "showInner",
            "showOuter",
            "showInnerBox",
            "showOuterBox",
        ]:
            if getattr(self.options, i):
                show_list.append(i.lower().replace("show", ""))
        opt = self.options
        ret = self.generate_nup(
            unit=opt.unit,
            pgSize=(opt.pgSizeX, opt.pgSizeY),
            pgMargin=(
                opt.pgMarginTop,
                opt.pgMarginRight,
                opt.pgMarginBottom,
                opt.pgMarginLeft,
            ),
            num=(opt.rows, opt.cols),
            calculateSize=opt.calculateSize,
            size=(opt.sizeX, opt.sizeY),
            margin=(opt.marginTop, opt.marginRight, opt.marginBottom, opt.marginLeft),
            padding=(
                opt.paddingTop,
                opt.paddingRight,
                opt.paddingBottom,
                opt.paddingLeft,
            ),
            show=show_list,
        )
        if ret:
            stream.write(ret)

    def expandTuple(self, unit, x, length=4):
        try:
            iter(x)
        except:
            return None

        if len(x) != length:
            x *= 2
        if len(x) != length:
            raise Exception("expandTuple: requires 2 or 4 item tuple")
        try:
            return tuple(
                map(
                    lambda ev: (
                        self.svg.unittouu(str(eval(str(ev))) + unit)
                        / self.svg.unittouu("1px")
                    ),
                    x,
                )
            )
        except:
            return None

    def generate_nup(
        self,
        unit="px",
        pgSize=("8.5*96", "11*96"),
        pgMargin=(0, 0),
        pgPadding=(0, 0),
        num=(2, 2),
        calculateSize=True,
        size=None,
        margin=(0, 0),
        padding=(20, 20),
        show=["default"],
    ):
        """Generate the SVG.  Inputs are run through 'eval(str(x))' so you can use
        '8.5*72' instead of 612.  Margin / padding dimension tuples can be
        (top & bottom, left & right) or (top, right, bottom, left).

        Keyword arguments:
        pgSize -- page size, width x height
        pgMargin -- extra space around each page
        pgPadding -- added to pgMargin
        n -- rows x cols
        size -- override calculated size, width x height
        margin -- white space around each piece
        padding -- inner padding for each piece
        show -- list of keywords indicating what to show
                - 'crosses' - cutting guides
                - 'inner' - inner boundary
                - 'outer' - outer boundary
        """

        if "default" in show:
            show = set(show).union(["inner", "innerbox", "holder", "crosses"])

        pgMargin = self.expandTuple(unit, pgMargin)
        pgPadding = self.expandTuple(unit, pgPadding)
        margin = self.expandTuple(unit, margin)
        padding = self.expandTuple(unit, padding)

        pgSize = self.expandTuple(unit, pgSize, length=2)
        #    num = tuple(map(lambda ev: eval(str(ev)), num))

        if not pgMargin or not pgPadding:
            return inkex.errormsg(_("No padding or margin available."))

        page_edge = list(map(sum, zip(pgMargin, pgPadding)))

        top, right, bottom, left = 0, 1, 2, 3
        width, height = 0, 1
        rows, cols = 0, 1
        size = self.expandTuple(unit, size, length=2)
        if (
            size is None
            or calculateSize
            or len(size) < 2
            or size[0] == 0
            or size[1] == 0
        ):
            size = (
                (
                    pgSize[width]
                    - page_edge[left]
                    - page_edge[right]
                    - num[cols] * (margin[left] + margin[right])
                )
                / num[cols],
                (
                    pgSize[height]
                    - page_edge[top]
                    - page_edge[bottom]
                    - num[rows] * (margin[top] + margin[bottom])
                )
                / num[rows],
            )

        # sep is separation between same points on pieces
        sep = (
            size[width] + margin[right] + margin[left],
            size[height] + margin[top] + margin[bottom],
        )

        style = "stroke:#000000;stroke-opacity:1;fill:none;fill-opacity:1;"

        padbox = Rectangle(
            x=str(page_edge[left] + margin[left] + padding[left]),
            y=str(page_edge[top] + margin[top] + padding[top]),
            width=str(size[width] - padding[left] - padding[right]),
            height=str(size[height] - padding[top] - padding[bottom]),
            style=style,
        )
        margbox = Rectangle(
            x=str(page_edge[left] + margin[left]),
            y=str(page_edge[top] + margin[top]),
            width=str(size[width]),
            height=str(size[height]),
            style=style,
        )

        doc = self.get_template(width=pgSize[width], height=pgSize[height])
        svg = doc.getroot()

        def make_clones(under, to):
            for row in range(0, num[rows]):
                for col in range(0, num[cols]):
                    if row == 0 and col == 0:
                        continue
                    use = under.add(Use())
                    use.set("xlink:href", "#" + to)
                    use.transform.add_translate(col * sep[width], row * sep[height])

        # guidelayer #####################################################
        if {"inner", "outer"}.intersection(show):
            layer = svg.add(inkex.Layer.new("Guide Layer"))
            if "inner" in show:
                ibox = layer.add(padbox.copy())
                ibox.style["stroke"] = "#8080ff"
                ibox.set("id", "innerguide")
                make_clones(layer, "innerguide")
            if "outer" in show:
                obox = layer.add(margbox.copy())
                obox.style["stroke"] = "#8080ff"
                obox.set("id", "outerguide")
                make_clones(layer, "outerguide")

        # crosslayer #####################################################
        if {"crosses"}.intersection(show):
            layer = svg.add(inkex.Layer.new("Cut Layer"))

            if "crosses" in show:
                crosslen = 12
                group = layer.add(inkex.Group(id="cross"))
                x, y = 0, 0
                path = "M%f %f" % (
                    x + page_edge[left] + margin[left],
                    y + page_edge[top] + margin[top] - crosslen,
                )
                path += " L%f %f" % (
                    x + page_edge[left] + margin[left],
                    y + page_edge[top] + margin[top] + crosslen,
                )
                path += " M%f %f" % (
                    x + page_edge[left] + margin[left] - crosslen,
                    y + page_edge[top] + margin[top],
                )
                path += " L%f %f" % (
                    x + page_edge[left] + margin[left] + crosslen,
                    y + page_edge[top] + margin[top],
                )
                group.add(
                    inkex.PathElement(
                        style=style + "stroke-width:0.05", d=path, id="crossmarker"
                    )
                )
                for row in 0, 1:
                    for col in 0, 1:
                        if row or col:
                            cln = group.add(Use())
                            cln.set("xlink:href", "#crossmarker")
                            cln.transform.add_translate(
                                col * size[width], row * size[height]
                            )
                make_clones(layer, "cross")

        # clonelayer #####################################################
        layer = svg.add(inkex.Layer.new("Clone Layer"))
        make_clones(layer, "main")

        # mainlayer ######################################################
        layer = svg.add(inkex.Layer.new("Main Layer"))
        group = layer.add(inkex.Group(id="main"))

        if "innerbox" in show:
            group.add(padbox)
        if "outerbox" in show:
            group.add(margbox)
        if "holder" in show:
            x, y = (
                page_edge[left] + margin[left] + padding[left],
                page_edge[top] + margin[top] + padding[top],
            )
            w, h = (
                size[width] - padding[left] - padding[right],
                size[height] - padding[top] - padding[bottom],
            )
            path = "M{:f} {:f}".format(x + w / 2.0, y)
            path += " L{:f} {:f}".format(x + w, y + h / 2.0)
            path += " L{:f} {:f}".format(x + w / 2.0, y + h)
            path += " L{:f} {:f}".format(x, y + h / 2.0)
            path += " Z"
            group.add(inkex.PathElement(style=style, d=path))

        return svg.tostring()


if __name__ == "__main__":
    Nup().run()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Manpreet Singh <manpreet.singh.dev@proton.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import math
from typing import Tuple

import inkex

MAX_CLIP_PATH_RECURSION = 1


class ClipPathRecursionError(Exception):
    pass


def interpolate_float(num1: float, num2: float, factor: float) -> float:
    return num1 + (num2 - num1) * factor


def safe_div(num1: float, num2: float, on_fail: float) -> float:
    return num1 / num2 if num2 != 0 else on_fail


def get_scale(tfm: inkex.Transform) -> Tuple[float, float]:
    dt = tfm.a * tfm.d - tfm.b * tfm.c

    if tfm.a or tfm.b:
        r = math.hypot(tfm.a, tfm.b)
        return r, dt / r

    if tfm.c or tfm.d:
        s = math.hypot(tfm.c, tfm.d)
        return dt / s, s

    return 0, 0


def get_scale_tfm(tfm: inkex.Transform) -> inkex.Transform:
    scale_x, scale_y = get_scale(tfm)
    return inkex.Transform((scale_x, 0, 0, scale_y, 0, 0))


def get_scale_tfm_abs(tfm: inkex.Transform) -> inkex.Transform:
    scale_x, scale_y = get_scale(tfm)
    return inkex.Transform((abs(scale_x), 0, 0, abs(scale_y), 0, 0))


def bbox_helper(values: Tuple[float, float, float, float]) -> inkex.BoundingBox:
    """Create a bounding box object from AF-sorted values"""
    return inkex.BoundingBox((values[0], values[2]), (values[1], values[3]))


def get_transformed_bbox(
    bbox: inkex.BoundingBox, tfm: inkex.Transform
) -> inkex.BoundingBox:
    """Get the bounding box of the transformed rectangle defined by `bbox` and `tfm`"""

    pts = [
        inkex.Vector2d(bbox.left, bbox.top),
        inkex.Vector2d(bbox.left, bbox.bottom),
        inkex.Vector2d(bbox.right, bbox.top),
        inkex.Vector2d(bbox.right, bbox.bottom),
    ]
    pts[:] = map(tfm.apply_to_point, pts)

    return inkex.BoundingBox(
        (min(pts, key=lambda p: p.x).x, max(pts, key=lambda p: p.x).x),
        (min(pts, key=lambda p: p.y).y, max(pts, key=lambda p: p.y).y),
    )



# coding=utf-8
#
# Copyright (C) 2007 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Python barcode renderer for Code39 Extended barcodes. Designed for Inkscape.
"""

from .Code39 import Code39

encode = list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")

charmap = {}

i = 0
for char in encode:
    charmap[char] = i
    i += 1


def get_map(array):
    """Extended encoding maps for full ASCII Code93"""
    result = {}
    for y, x in enumerate(array):
        result[chr(x)] = encode[y]

    return result


# MapA is eclectic, but B, C, D are all ASCII ranges
mapA = get_map(
    [
        27,
        28,
        29,
        30,
        31,
        59,
        60,
        61,
        62,
        63,
        91,
        92,
        93,
        94,
        95,
        123,
        124,
        125,
        126,
        127,
        0,
        64,
        96,
        127,
        127,
        127,
    ]
)  # %
mapB = get_map(range(1, 26))  # $
mapC = get_map(range(33, 58))  # /
mapD = get_map(range(97, 122))  # +


class Code39Ext(Code39):
    """Render Code39 Barcodes"""

    def encode(self, text):
        # We are only going to extend the Code39 barcodes
        result = ""
        for character in text:
            if character in mapA:
                character = "%" + mapA[character]
            elif character in mapB:
                character = "$" + mapB[character]
            elif character in mapC:
                character = "/" + mapC[character]
            elif character in mapD:
                character = "+" + mapD[character]
            result = result + character

        return Code39.encode(self, result)



# coding=utf-8
#
# Copyright (C) 2010 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
Base module for rendering barcodes for Inkscape.
"""

import itertools
import sys

from inkex import Group, TextElement, Rectangle
from inkex.localization import inkex_gettext as _

(TEXT_POS_BOTTOM, TEXT_POS_TOP) = range(2)
(WHITE_BAR, BLACK_BAR, TALL_BAR) = range(3)
TEXT_TEMPLATE = "font-size:%dpx;text-align:center;text-anchor:middle;"

try:
    from typing import Optional
except ImportError:
    pass


class Barcode(object):
    """Provide a base class for all barcode renderers"""

    default_height = 30
    font_size = 9
    name = None  # type: Optional[str]

    def error(self, text, msg):
        """Cause an error to be reported"""
        sys.stderr.write(
            _("Error encoding '{}' as {} barcode: {}\n").format(text, self.name, msg)
        )
        return "ERROR"

    def encode(self, text):
        """
        Replace this with the encoding function, it should return
        a string of ones and zeros
        """
        raise NotImplementedError("You need to write an encode() function.")

    def __init__(self, param):
        param = param or {}
        self.document = param.get("document", None)
        self.known_ids = []
        self._extra = []

        self.pos_x = int(param.get("x", 0))
        self.pos_y = int(param.get("y", 0))
        self.text = param.get("text", None)
        self.scale = param.get("scale", 1)
        self.height = param.get("height", self.default_height)
        self.pos_text = param.get("text_pos", TEXT_POS_BOTTOM)

        if self.document:
            self.known_ids = list(self.document.xpath("//@id"))

        if not self.text:
            raise ValueError(_("No string specified for barcode."))

    def get_id(self, name="element"):
        """Get the next useful id (and claim it)"""
        index = 0
        while name in self.known_ids:
            index += 1
            name = "barcode{:d}".format(index)
        self.known_ids.append(name)
        return name

    def add_extra_barcode(self, barcode, **kw):
        """Add an extra barcode along side this one, used for ean13 extras"""
        from . import get_barcode

        kw["height"] = self.height
        kw["document"] = self.document
        kw["scale"] = None
        self._extra.append(get_barcode(barcode, **kw).generate())

    def generate(self):
        """Generate the actual svg from the coding"""
        string = self.encode(self.text)

        if string == "ERROR":
            return

        name = self.get_id("barcode")

        # use an svg group element to contain the barcode
        barcode = Group()
        barcode.set("id", name)
        barcode.set("style", "fill: black;")

        barcode.transform.add_translate(self.pos_x, self.pos_y)
        if self.scale:
            barcode.transform.add_scale(self.scale)

        bar_id = 1
        bar_offset = 0
        tops = set()

        for datum in self.graphical_array(string):
            # Datum 0 tells us what style of bar is to come next
            style = self.get_style(int(datum[0]))
            # Datum 1 tells us what width in units,
            # style tells us how wide a unit is
            width = int(datum[1]) * int(style["width"])

            if style["write"]:
                tops.add(style["top"])
                rect = Rectangle()
                rect.set("x", str(bar_offset))
                rect.set("y", str(style["top"]))
                if self.pos_text == TEXT_POS_TOP:
                    rect.set("y", str(style["top"] + self.font_size))
                rect.set("id", "{}_bar{:d}".format(name, bar_id))
                rect.set("width", str(width))
                rect.set("height", str(style["height"]))
                barcode.append(rect)
            bar_offset += width
            bar_id += 1

        for extra in self._extra:
            if extra is not None:
                barcode.append(extra)

        bar_width = bar_offset
        # Add text at the bottom of the barcode
        text = TextElement()
        text.set("x", str(int(bar_width / 2)))
        text.set("y", str(min(tops) + self.font_size - 1))
        if self.pos_text == TEXT_POS_BOTTOM:
            text.set("y", str(self.height + max(tops) + self.font_size))
        text.set("style", TEXT_TEMPLATE % self.font_size)
        text.set("xml:space", "preserve")
        text.set("id", "{}_text".format(name))
        text.text = str(self.text)
        barcode.append(text)
        return barcode

    def graphical_array(self, code):
        """Converts black and white markets into a space array"""
        return [(x, len(list(y))) for x, y in itertools.groupby(code)]

    def get_style(self, index):
        """Returns the styles that should be applied to each bar"""
        result = {"width": 1, "top": 0, "write": True}
        if index == BLACK_BAR:
            result["height"] = int(self.height)
        if index == TALL_BAR:
            result["height"] = int(self.height) + int(self.font_size / 2)
        if index == WHITE_BAR:
            result["write"] = False
        return result



#!/usr/bin/env python3

# Copyright (C) 2024 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Parses the extracted contents of an afdesign file
Based on https://github.com/VMDevCpp/afread, MIT-licensed
"""

from __future__ import annotations
from io import BufferedIOBase
from typing import Optional, List, Dict, Union

from .enhance import process
from .utils import BinaryParser, BitReader
from .consts import content_magic
from .types import (
    Document,
    ObjectStatus,
    AFObjectMetadata,
    EmbeddedData,
    MetadataStatus,
    AFObject,
    EnumT,
    AFListObject,
    AFDictObject,
    FlagsT,
    Field,
    Curve12,
    Curve16,
    Curve18,
    Curve18Type,
    UnknownStruct,
)


class AFParser(BinaryParser):
    data: BitReader

    def __init__(self, data: BufferedIOBase):
        super().__init__(BitReader(data))
        self.document = Document(0, [], ObjectStatus.ROOT, {}, {})

        self.reader_map = {
            0x01: self.read_u8,  # uint8_t
            0x02: self.read_u16,  # uint16_t
            0x03: self.read_u32,  # uint32_t
            0x04: lambda: int.from_bytes(
                self.read(8), byteorder="little", signed=False
            ),  # uint64_t
            0x05: lambda: int.from_bytes(
                self.read(1), byteorder="little", signed=True
            ),  # int8_t
            0x06: lambda: int.from_bytes(
                self.read(2), byteorder="little", signed=True
            ),  # int16_t
            0x07: self.read_i32,  # int32_t
            0x08: lambda: int.from_bytes(
                self.read(8), byteorder="little", signed=True
            ),  # int64_t
            0x09: self.read_float,  # float
            0x0A: self.read_double,  # double
            0x15: lambda: [self.read_i32() for _ in range(2)],
            0x16: lambda: [self.read_i32() for _ in range(3)],
            0x17: lambda: [self.read_i32() for _ in range(4)],
            0x18: lambda: [self.read_i32() for _ in range(5)],
            0x19: lambda: [self.read_i32() for _ in range(6)],
            0x1F: lambda: [self.read_float() for _ in range(2)],
            0x20: lambda: [self.read_float() for _ in range(3)],
            0x21: lambda: [self.read_float() for _ in range(4)],
            0x22: lambda: [self.read_float() for _ in range(5)],
            0x23: lambda: [self.read_float() for _ in range(6)],
            0x24: lambda: [self.read_double() for _ in range(2)],
            0x25: lambda: [self.read_double() for _ in range(3)],
            0x26: lambda: [self.read_double() for _ in range(4)],
            0x27: lambda: [self.read_double() for _ in range(5)],
            0x28: lambda: [self.read_double() for _ in range(6)],
            0x29: self.data.read_bit,
            0x2B: self.read_string,
            0x2D: self.read_binary_data,
            0x2E: self.read_string,
            0x2F: self.read_u32,
            0x30: self.load_object_30,
            0x31: self.load_object_31,
            0x33: self.read_embedded_data,
            0x34: self.read_u32,
            0x75: self.read_flagsT,
        }
        # These are only for debugging the parser.
        self.tags: List[Dict[str, Union[int, str]]] = []

    def add_tag(self, caption: str, start: int, end: int, color: Optional[str] = None):
        self.tags.append({"start": start, "end": end, "caption": caption})

    def set_end_tag(self, start: int, end: int):
        [i for i in self.tags if i["start"] == start][0]["end"] = end

    def parse(self) -> Document:
        df = self.document.doc_format

        df["file_tag"] = self.read_u32()
        assert df["file_tag"] == content_magic

        df["file_ver"] = self.read_u16()
        df["type_tag"] = self.read_tag()
        df["type_ver"] = self.read_u16()

        if df["file_ver"] > 2:
            raise Exception("Unknown file format")
        elif df["file_ver"] < 2:
            df["version"] = 0
        else:
            df["version"] = self.read_u32()

        self.document.id = 0
        self.document.types = [AFObjectMetadata(0, df["file_tag"], MetadataStatus.TAG)]

        self.load_fields(self.document, True)
        # Postprocessing step
        return process(self.document)

    def format_position(self):
        return f"{self.get_pos()} (line {hex(self.get_pos() // 16 * 16)}, col {hex(self.get_pos() % 16)})"

    def load_fields(self, parent: AFObject, with_tag: bool = True):
        """Load all following fields into parent"""
        while True:
            pos = self.get_pos()
            type_ = self.read_u8()
            array = type_ >= 128
            type_ &= 0x7F

            if type_ == 0x00:
                return
            if type_ > 0x77:
                raise ValueError(f"invalid type at {pos}")

            tag = self.read_tag() if with_tag else None

            need_tag = type_ in (0x30, 0x31, 0x32, 0x33, 0x75, 0x76, 0x77)
            if need_tag and tag is None:
                raise ValueError("missing tag")

            result = Field(type_, self.load_field(type_, array), parent=parent)

            parent.append(tag, result)

    def load_field(self, type_: int, array: bool):
        count = 1
        if array:
            if type_ in (0x2B, 0x2E):
                _ = self.read_u32()  # total_size comes before count
            if type_ in (0x2D, 0x33, 0x75):
                raise ValueError(
                    "Don't know how arrays for embedded/binary/flagsT data work"
                )
            count = self.read_u32()

        if type_ == 0x29:
            self.data.clear_bits()

        # Get loader function
        load_func = None
        if type_ in self.reader_map:
            load_func = self.reader_map[type_]
        elif type_ == 0x2A:
            version = self.read_u16() if array else None

            def load_func():
                return self.read_enumT(version)
        elif type_ == 0x2C:
            size = self.read_u16()

            def load_func():
                return self.read_curve_struct(size)
        elif type_ == 0x32:
            h32 = None
            if array:
                tag2 = self.read_tag()
                h32 = AFObjectMetadata(self.read_u16(), tag2, MetadataStatus.TAG_ID)

            def load_func():
                return self.load_object_32(h32)

        elif 0x35 <= type_ <= 0x74:

            def load_func():
                return self.read_struct(type_)

        if load_func is not None:
            tmp = [load_func() for _ in range(count)]
            return tmp[0] if not array else tmp
        else:
            raise ValueError(
                f"Unknown type {hex(type_)} at position {self.format_position()}."
            )

    def read_string(self):
        length = self.read_u32()
        return self.read(length).decode("utf-8")

    def read_enumT(self, version=None):
        return EnumT(self.read_u16(), self.read_u16() if version is None else version)

    def read_curve_struct(self, size: int):
        if size == 12:
            return Curve12(
                self.read_double(),
                (self.read_u8(), self.read_u8(), self.read_u8(), self.read_u8()),
            )

        elif size == 16:
            return Curve16(self.read_double(), (self.read_u32(), self.read_u32()))
        elif size == 18:
            return Curve18(
                (self.read_double(), self.read_double()),
                Curve18Type(self.read_u8()),
                self.read_u8(),
            )

        elif size in [24, 32]:
            return UnknownStruct(self.read(size))

        else:
            raise ValueError("Bad size for curve struct")

    def read_struct(self, type_: int):
        # We only read the data here since would otherwise need parent properties
        # to discriminate further. This is done in a later step, in enhance.py
        size = type_ - 0x34
        return UnknownStruct(self.read(size))

    def read_embedded_data(self):
        tag = self.read_tag()
        size = self.read_u32()
        data = self.read(size)
        for char in data:
            if not (chr(char).isalnum() or chr(char) == "/"):
                raise ValueError(
                    f"Invalid embedded data, encountered character {chr(char)}"
                )
        return EmbeddedData(tag, size, data.decode("latin-1"))

    def read_binary_data(self):
        size = self.read_u32()
        return self.read(size)

    def read_flagsT(self):
        version = self.read_u16()
        count = self.read_u8()
        flags = self.read(count)
        return FlagsT(version, flags)

    def load_object_30(self) -> AFListObject:
        """Object of type 30"""
        dest = AFListObject(0, [], ObjectStatus.NON_SHARED, [])
        self.load_fields(dest, False)
        return dest

    def load_object_31(self) -> AFDictObject:
        """Object of type 31"""
        dest = AFDictObject(0, [], ObjectStatus(self.read_u8()), {})
        if dest.status == ObjectStatus.SHARED:
            dest.id = self.read_u32()
            # TODO this ID needs to be unique

            while True:
                status = AFObjectMetadata(0, "", MetadataStatus.TAG)
                dest.types.append(status)
                type_flag = self.read_u8()
                if type_flag == 0:
                    status.tag = self.read_tag()
                    status.id = self.read_u16()
                    self.load_fields(dest)
                    status.status = MetadataStatus.TAG_ID
                elif type_flag == 1:
                    status.tag = self.read_tag()
                    break
                elif type_flag == 2:
                    dest.types.remove(status)
                    break
                else:
                    raise ValueError(
                        f"Unknown status type flag at {self.format_position()}"
                    )

            self.load_fields(dest)
        elif dest.status == ObjectStatus.LINK:
            dest.id = self.read_u32()
            # TODO this ID needs to be unique

        elif dest.status == ObjectStatus.NULL:
            pass
        else:
            raise ValueError(
                f"Unknown status flag {dest.status} at {self.format_position()}"
            )
        return dest

    def load_object_32(self, status: Optional[AFObjectMetadata] = None) -> AFDictObject:
        flag = self.read_u8()
        dest = AFDictObject(
            0, [], ObjectStatus.NULL if flag == 0 else ObjectStatus.NON_SHARED, {}
        )
        if dest.status == ObjectStatus.NON_SHARED:
            dest.id = 0
            if status is None:
                tag = self.read_tag()
                status = AFObjectMetadata(self.read_u16(), tag, MetadataStatus.TAG_ID)
            dest.types.append(status)

            self.load_fields(dest)
        return dest



#!/usr/bin/env python3
# coding=utf-8

import os
from inkex import load_svg, TemplateExtension


class ShapeBuilderTemplate(TemplateExtension):
    """Generate shape builder pattern"""

    def add_arguments(self, pars):
        pars.add_argument("--svg", help="Template to load")

    def get_size(self):
        return (900, "px", 900, "px")

    def get_template(self, **kwargs):
        path = os.path.dirname(os.path.realpath(__file__))
        return load_svg(os.path.join(path, self.options.svg))

    def set_namedview(self, width_px, height_px, unit):
        pass


if __name__ == "__main__":
    ShapeBuilderTemplate().run()



#!/usr/bin/env python3
"""To black and white"""

import inkex


class BlackAndWhite(inkex.ColorExtension):
    """Convert colours to black and white"""

    def add_arguments(self, pars):
        pars.add_argument(
            "-t", "--threshold", type=int, default=127, help="Threshold Color Level"
        )

    def modify_color(self, name, color):
        # ITU-R Recommendation BT.709 (NTSC and PAL)
        # l = 0.2125 * r + 0.7154 * g + 0.0721 * b
        lum = 0.299 * color.red + 0.587 * color.green + 0.114 * color.blue
        grey = 255 if lum > self.options.threshold else 0
        return inkex.Color((grey, grey, grey))


if __name__ == "__main__":
    BlackAndWhite().run()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Martin Owens <doctormo@gmail.com>
#                    Sergei Izmailov <sergei.a.izmailov@gmail.com>
#                    Ryan Jarvis <ryan@shopboxretail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=arguments-differ
"""
Interface for all group based elements such as Groups, Use, Markers etc.
"""

from lxml import etree  # pylint: disable=unused-import

from ..paths import Path
from ..transforms import BoundingBox, Transform

from ._utils import addNS
from ._base import ShapeElement, ViewboxMixin
from ._polygons import PathElement

try:
    from typing import Optional, List  # pylint: disable=unused-import
except ImportError:
    pass


class GroupBase(ShapeElement):
    """Base Group element"""

    def get_path(self):
        ret = Path()
        for child in self:
            if isinstance(child, ShapeElement):
                ret += child.path.transform(child.transform)
        return ret

    def bounding_box(self, transform=None):
        # type: (Optional[Transform]) -> Optional[BoundingBox]
        """BoundingBox of the shape

        .. versionchanged:: 1.4
            Exclude invisible child objects from bounding box computation

        .. versionchanged:: 1.1
            result adjusted for element's clip path if applicable.
        """
        bbox = None
        effective_transform = Transform(transform) @ self.transform
        for child in self:
            if isinstance(child, ShapeElement) and child.is_visible():
                child_bbox = child.bounding_box(transform=effective_transform)
                if child_bbox is not None:
                    bbox += child_bbox
        clip = self.clip
        if clip is None or bbox is None:
            return bbox
        return bbox & clip.bounding_box(Transform(transform) @ self.transform)

    def shape_box(self, transform=None):
        # type: (Optional[Transform]) -> Optional[BoundingBox]
        """BoundingBox of the unclipped shape

        .. versionchanged:: 1.4
            returns the bounding box without possible clip effects of child objects

        .. versionadded:: 1.1
            Previous :func:`bounding_box` function, returning the bounding box
            without computing the effect of a possible clip.
        """
        bbox = None
        effective_transform = Transform(transform) @ self.transform
        for child in self:
            if isinstance(child, ShapeElement):
                child_bbox = child.shape_box(transform=effective_transform)
                if child_bbox is not None:
                    bbox += child_bbox
        return bbox

    def bake_transforms_recursively(self, apply_to_paths=True):
        """Bake transforms, i.e. each leaf node has the effective transform (starting
        from this group) set, and parent transforms are removed.

        .. versionadded:: 1.4

        Args:
            apply_to_paths (bool, optional): For path elements, the
                path data is transformed with its effective transform. Nodes and handles
                will have the same position as before, but visual appearance of the
                stroke may change (stroke-width is not touched). Defaults to True.
        """
        # pylint: disable=attribute-defined-outside-init
        self.transform: Transform
        for element in self:
            if isinstance(element, PathElement) and apply_to_paths:
                element.path = element.path.transform(self.transform)
            else:
                element.transform = self.transform @ element.transform
                if isinstance(element, GroupBase):
                    element.bake_transforms_recursively(apply_to_paths)
        self.transform = None


class Group(GroupBase):
    """Any group element (layer or regular group)"""

    tag_name = "g"

    @classmethod
    def new(cls, label, *children, **attrs):
        attrs["inkscape:label"] = label
        return super().new(*children, **attrs)

    def effective_style(self):
        """A blend of each child's style mixed together (last child wins)"""
        style = self.style
        for child in self:
            style.update(child.effective_style())
        return style

    @property
    def groupmode(self):
        """Return the type of group this is"""
        return self.get("inkscape:groupmode", "group")


class Layer(Group):
    """Inkscape extension of svg:g"""

    def _init(self):
        self.set("inkscape:groupmode", "layer")

    @classmethod
    def is_class_element(cls, elem):
        # type: (etree.Element) -> bool
        return (
            elem.get("{http://www.inkscape.org/namespaces/inkscape}groupmode", None)
            == "layer"
        )


class Anchor(GroupBase):
    """An anchor or link tag"""

    tag_name = "a"

    @classmethod
    def new(cls, href, *children, **attrs):
        attrs["xlink:href"] = href
        return super().new(*children, **attrs)


class ClipPath(GroupBase):
    """A path used to clip objects"""

    tag_name = "clipPath"


class Marker(GroupBase, ViewboxMixin):
    """The <marker> element defines the graphic that is to be used for drawing
    arrowheads or polymarkers on a given <path>, <line>, <polyline> or <polygon>
    element."""

    tag_name = "marker"

    def get_viewbox(self) -> List[float]:
        """Returns the viewbox of the Marker, falling back to
        [0 0 markerWidth markerHeight]

        .. versionadded:: 1.3"""
        vbox = self.get("viewBox", None)
        result = self.parse_viewbox(vbox)
        if result is None:
            # use viewport, https://www.w3.org/TR/SVG11/painting.html#MarkerElement
            return [
                0,
                0,
                self.to_dimensionless(self.get("markerWidth")),
                self.to_dimensionless(self.get("markerHeight")),
            ]
        return result



#!/usr/bin/env python3
"""Convert text to upper case"""

import inkex


class Uppercase(inkex.TextExtension):
    """To upper case"""

    def process_chardata(self, text):
        return text.upper()


if __name__ == "__main__":
    Uppercase().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2013 <Public Domain>
#               2018 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import print_function

import inkex
from inkex.localization import inkex_gettext as _

import hpgl_encoder


class HpglOutput(inkex.OutputExtension):
    """Save as HPGL Output"""

    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument(
            "--resolutionX", type=float, default=1016.0, help="Resolution X (dpi)"
        )
        pars.add_argument(
            "--resolutionY", type=float, default=1016.0, help="Resolution Y (dpi)"
        )
        pars.add_argument("--pen", type=int, default=1, help="Pen number")
        pars.add_argument("--force", type=int, default=0, help="Pen force (g)")
        pars.add_argument("--speed", type=int, default=0, help="Pen speed (cm/s)")
        pars.add_argument(
            "--orientation", type=str, default="0", help="Rotation (Clockwise)"
        )
        pars.add_argument(
            "--mirrorX", type=inkex.Boolean, default=False, help="Mirror X axis"
        )
        pars.add_argument(
            "--mirrorY", type=inkex.Boolean, default=False, help="Mirror Y axis"
        )
        pars.add_argument(
            "--center", type=inkex.Boolean, default=False, help="Center zero point"
        )
        pars.add_argument("--overcut", type=float, default=1.0, help="Overcut (mm)")
        pars.add_argument(
            "--precut", type=inkex.Boolean, default=True, help="Use precut"
        )
        pars.add_argument("--flat", type=float, default=1.2, help="Curve flatness")
        pars.add_argument(
            "--autoAlign", type=inkex.Boolean, default=True, help="Auto align"
        )
        pars.add_argument(
            "--toolOffset",
            type=float,
            default=0.25,
            help="Tool (Knife) offset correction (mm)",
        )

    def save(self, stream):
        self.options.debug = False
        # get hpgl data
        if len(self.svg.xpath("//svg:use|//svg:flowRoot|//svg:text")) > 0:
            self.preprocess(["flowRoot", "text"])
        encoder = hpgl_encoder.hpglEncoder(self)
        try:
            hpgl = encoder.getHpgl()
        except hpgl_encoder.NoPathError:
            raise inkex.AbortExtension(_("No convertible objects were found"))
        # convert raw HPGL to HPGL
        hpgl_init = "IN"
        # if self.options.force > 0:
        #    hpgl_init += ';FS%d' % self.options.force
        # if self.options.speed > 0:
        #    hpgl_init += ';VS%d' % self.options.speed
        hpgl = hpgl_init + hpgl + ";SP0;PU0,0;IN; "
        stream.write(hpgl.encode("utf-8"))


if __name__ == "__main__":
    HpglOutput().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Twirl path extension (modify path)"""

import math
import inkex


class Twirl(inkex.EffectExtension):
    """Modify a path by twisting the nodes around a point"""

    def add_arguments(self, pars):
        pars.add_argument(
            "-t", "--twirl", type=float, default=5.0, help="amount of twirl"
        )
        pars.add_argument(
            "-r",
            "--rotation",
            type=inkex.Boolean,
            default=True,
            help="direction of rotation",
        )

    def effect(self):
        view_center = self.svg.namedview.center
        rotation = 1 if self.options.rotation else -1
        twirl = self.options.twirl / 1000
        for node in self.svg.selection.filter(inkex.PathElement):
            self.twirl_node(view_center, rotation, twirl, node)

    @staticmethod
    def twirl_node(center, direction, ammount, node):
        """Apply a twirl to a path given the center, direction and amount"""
        path = node.path.to_superpath()
        for sub in path:
            for csp in sub:
                for point in csp:
                    point[0] -= center[0]
                    point[1] -= center[1]
                    dist = math.sqrt((point[0] ** 2) + (point[1] ** 2))
                    if dist != 0:
                        art = direction * dist * ammount
                        theta = math.atan2(point[1], point[0]) + art
                        point[0] = dist * math.cos(theta)
                        point[1] = dist * math.sin(theta)
                    point[0] += center[0]
                    point[1] += center[1]
        node.path = path


if __name__ == "__main__":
    Twirl().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#               2020 Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Interpolate between two paths, respecting their style"""

import copy

import inkex
from inkex.utils import pairwise
from inkex.tween import (
    AttributeInterpolator,
    StyleInterpolator,
    EqualSubsegmentsInterpolator,
    FirstNodesInterpolator,
)

from inkex.localization import inkex_gettext as _


class Interpolate(inkex.EffectExtension):
    """Interpolate extension"""

    def add_arguments(self, pars):
        pars.add_argument(
            "-e",
            "--exponent",
            type=float,
            default=1.0,
            help="values other than zero give non linear interpolation",
        )
        pars.add_argument(
            "-s", "--steps", type=int, default=5, help="number of interpolation steps"
        )
        pars.add_argument(
            "-m",
            "--method",
            type=str,
            default="equalSubsegments",
            help="method of interpolation",
        )
        pars.add_argument(
            "-d", "--dup", type=inkex.Boolean, default=True, help="duplicate endpaths"
        )
        pars.add_argument(
            "--style",
            type=inkex.Boolean,
            default=False,
            help="try interpolation of some style properties",
        )
        pars.add_argument(
            "--zsort",
            type=inkex.Boolean,
            default=False,
            help="use z-order instead of selection order",
        )

    def effect(self):
        steps = self.get_steps()

        objectpairs = self.get_copied_path_pairs()
        if not objectpairs:
            raise inkex.AbortExtension(_("At least two paths need to be selected"))

        for elem1, elem2 in objectpairs:
            method = EqualSubsegmentsInterpolator
            if self.options.method == "firstNodes":
                method = FirstNodesInterpolator

            path_interpolator = AttributeInterpolator.create_from_attribute(
                elem1, elem2, "d", method=method
            )
            if self.options.style:
                style_interpolator = StyleInterpolator(elem1, elem2)

            group = self.svg.get_current_layer().add(inkex.Group())
            for time in steps:
                interpolated_path = path_interpolator.interpolate(time)
                new = group.add(inkex.PathElement())
                new.path = interpolated_path
                if self.options.style:
                    interpolated_style = style_interpolator.interpolate(time)
                    new.style = interpolated_style
                else:
                    new.style = copy.deepcopy(elem1.style)

    def get_steps(self):
        """Returns the interpolation steps as a monotonous array with elements between 0 and 1.
        0 and 1 are added as first and last elements if the source paths should be duplicated
        """
        exponent = self.options.exponent
        # if exponent >= 0:
        #    exponent += 1.0
        # else:
        #    exponent = 1.0 / (1.0 - exponent)
        steps = [
            ((i + 1) / (self.options.steps + 1.0)) ** exponent
            for i in range(self.options.steps)
        ]
        if self.options.dup:
            steps = [0] + steps + [1]
        return steps

    def get_copied_path_pairs(self):
        """Returns deep copies of pairs of subsequent objects of the
        current selection, either in z order or in selection order.
        Objects other than path objects are ignored. Transforms are baked in."""
        if self.options.zsort:
            # work around selection order swapping with Live Preview
            objects = self.svg.selection.rendering_order()
        else:
            # use selection order (default)
            objects = self.svg.selection

        objects = [
            node for node in objects.values() if isinstance(node, inkex.PathElement)
        ]

        for node in objects:
            node.apply_transform()

        objectpairs = pairwise(objects, start=False)
        return objectpairs


if __name__ == "__main__":
    Interpolate().run()



"""
inkvn Converter

Convert the intermediate data to Inkscape read by read.py
"""

import logging
from typing import List, Optional, Tuple, Union

import inkex
from inkex.base import SvgOutputMixin
import lxml.etree

from ..reader.read import CurveReader

from ..elements.artboard import VNArtboard
from ..elements.base import VNBaseElement
from ..elements.guide import VNGuideElement
from ..elements.group import VNGroupElement
from ..elements.image import VNImageElement
from ..elements.path import VNPathElement
from ..elements.text import VNTextElement, singleStyledText
from ..elements.styles import VNColor, VNGradient, pathStrokeStyle


class CurveConverter:
    """
    inkvn CurveConverter

    Convert the intermediate data to Inkscape.
    """

    def __init__(self) -> None:
        self.reader: CurveReader
        self.target_version: int
        self.has_transform_applied: bool
        self.doc: lxml.etree._ElementTree
        self.document: inkex.SvgDocumentElement
        self.offset_x: float
        self.offset_y: float

    def convert(self, reader: CurveReader, clip_page: bool = False) -> None:
        self.reader = reader

        """
        file version check

        New curve holds path data without transforms
        However, old curve has transforms already applied.
        I don't know exactly when the behavior has changed, so it's set to 30 as placeholder
        """
        self.has_transform_applied = reader.file_version < 30

        first_artboard = reader.artboards[0]

        self.doc = SvgOutputMixin.get_template(
            width=first_artboard.frame.width,
            height=first_artboard.frame.height,
        )
        self.document = self.doc.getroot()

        # Adding comments
        comment = lxml.etree.Comment(" Converted by extension-curve ")
        self.document.addprevious(comment)

        # first artboard becomes the front page
        # other artboards will be placed relative to the first one
        self.offset_x = first_artboard.frame.x
        self.offset_y = first_artboard.frame.y

        for target_artboard in reader.artboards:
            page = inkex.Page.new(
                width=target_artboard.frame.width,
                height=target_artboard.frame.height,
                x=target_artboard.frame.x - self.offset_x,
                y=target_artboard.frame.y - self.offset_y,
            )
            self.document.namedview.add(page)
            page.set("inkscape:label", target_artboard.title)

            self.load_page(
                self.document.add(inkex.Layer.new(label=target_artboard.title)),
                target_artboard,
                clip_page,
            )
            self.doc.getroot()

    def load_page(
        self, root_layer: inkex.Layer, artboard: VNArtboard, clip_page: bool = False
    ) -> None:
        """Convert  VNArtboard to inkex page."""

        # artboards have translations
        tr = inkex.transforms.Transform()
        tr_vector = inkex.Vector2d(
            artboard.frame.x - self.offset_x, artboard.frame.y - self.offset_y
        )
        tr.add_translate(tr_vector)
        root_layer.transform = tr

        # Artboard color/gradient
        rect = inkex.Rectangle.new(0, 0, artboard.frame.width, artboard.frame.height)
        rect.label = "background"

        # Fill Style
        if artboard.fillColor is not None:
            self.set_fill_color_styles(rect, artboard.fillColor)
            root_layer.add(rect)
        elif artboard.fillGradient is not None:
            if artboard.fillGradient.transform is not None:
                artboard.fillGradient.gradient.set(
                    "gradientTransform", artboard.fillGradient.transform
                )
            self.set_fill_grad_styles(rect, artboard.fillGradient)
            root_layer.add(rect)
        # if fill is none, rect will be dismissed

        # artboard clipping (optional), similar to Vectornator svg output.
        if clip_page:
            clip = inkex.ClipPath()
            clip_rect = inkex.Rectangle.new(
                0, 0, artboard.frame.width, artboard.frame.height
            )
            clip_rect.label = "page clipping"

            clip.add(clip_rect)
            self.document.defs.add(clip)

            if clip is not None:
                root_layer.style["clip-path"] = f"url(#{clip.get_id()})"

        # layers in the artboard
        for layer in artboard.layers:
            parent = root_layer.add(inkex.Layer.new(layer.name))
            parent.set("opacity", layer.opacity)
            parent.style["display"] = "inline" if layer.isVisible else "none"
            if layer.isLocked:
                parent.set("sodipodi:insensitive", "true")
            # ? isExpanded

            # elements in the layer
            for element in layer.elements:
                elm = self.load_element(element)
                if elm is not None:
                    parent.add(elm)

        # Guide
        for guide in artboard.guides:
            if guide is not None:
                self.add_guide(guide, tr_vector)

    def load_element(self, element: VNBaseElement) -> Optional[inkex.BaseElement]:
        """Converts an element to an SVG element."""
        try:
            if isinstance(element, VNGroupElement):
                return self.convert_group(element)

            elif isinstance(element, VNImageElement):
                return self.convert_image(element)

            elif isinstance(element, VNPathElement):
                return self.convert_path(element)

            elif isinstance(element, VNTextElement):
                # inkex.utils.debug(f'{element.name}: Text has been successfully parsed, but text import is not supported yet.')
                return self.convert_text(element)

            elif isinstance(element, VNBaseElement):
                return self.convert_base(element)  # will be empty path element

        except Exception as e:
            logging.error(f"Error converting element: {e}", exc_info=True)

    def convert_group(self, group_element: VNGroupElement) -> inkex.Group:
        """
        Converts a VNGroupElement to an SVG group (inkex.Group),
        or ClipPath when the group contains clipping mask.
        """

        group = inkex.Group()

        # BaseElement
        group.label = group_element.name
        group.style["opacity"] = group_element.opacity
        group.style["mix-blend-mode"] = group_element.convert_blend()
        group.style["display"] = "none" if group_element.isHidden else "inline"
        if group_element.isLocked:
            group.set("sodipodi:insensitive", "true")
        if not self.has_transform_applied and group_element.localTransform is not None:
            group.transform = group_element.localTransform.convert_transform()

        # blur
        if group_element.blur > 0:
            self.set_blur(group, group_element.convert_blur())

        # clipping mask
        clip_path_child = None
        clip = None
        for child in group_element.groupElements:
            # find the clipping mask
            if isinstance(child, VNPathElement) and child.mask == 1:
                clip_path_child = child
                break

        if clip_path_child:
            clip = inkex.ClipPath()
            clip_path_element = self.convert_path(clip_path_child)

            clip.add(clip_path_element)
            self.document.defs.add(clip)

        if clip is not None:
            group.style["clip-path"] = f"url(#{clip.get_id()})"

        for child in group_element.groupElements:
            svg_element = self.load_element(child)
            if svg_element is not None:
                group.add(svg_element)

        return group

    def convert_image(
        self, image_element: VNImageElement
    ) -> Union[inkex.Image, inkex.Group]:
        """Converts a VNImageElement to an SVG image (inkex.Image)."""
        image = inkex.Image()

        # BaseElement
        image.label = image_element.name
        image.style["opacity"] = image_element.opacity
        image.style["mix-blend-mode"] = image_element.convert_blend()
        image.style["display"] = "none" if image_element.isHidden else "inline"
        if image_element.isLocked:
            image.set("sodipodi:insensitive", "true")
        if image_element.transform is not None:
            image.transform = image_element.transform
        elif (
            not self.has_transform_applied and image_element.localTransform is not None
        ):
            image.transform = image_element.localTransform.convert_transform()

        # blur
        if image_element.blur > 0:
            self.set_blur(image, image_element.convert_blur())

        # Image
        img_format = image_element.image_format()
        width, height = image_element.image_dimension()

        image.set("preserveAspectRatio", "none")
        image.set(
            inkex.addNS("href", "xlink"),
            f"data:image/{img_format.lower()};base64,{image_element.imageData}",
        )
        image.set("width", width)
        image.set("height", height)

        # image cropping with clipping mask
        if image_element.cropRect is not None:
            clip = inkex.ClipPath()
            clip_element = image_element.convert_crop_rect()

            clip.add(clip_element)
            self.document.defs.add(clip)
            image.style["clip-path"] = f"url(#{clip.get_id()})"

        return image

    def convert_path(self, path_element: VNPathElement) -> inkex.PathElement:
        """Converts a VNPathElement to an SVG path (inkex.PathElement)."""
        path = inkex.PathElement()

        # pathGeometry
        if path_element.pathGeometries:
            for path_geometry in path_element.pathGeometries:
                path.path += path_geometry.path

        if not self.has_transform_applied and path_element.localTransform is not None:
            path.transform = path_element.localTransform.convert_transform()

            # Apply Transform
            path = inkex.PathElement.new(path.get_path().transform(path.transform))

        # PathEffect(corner), does not work for other paths in compoundPath
        if (
            not self.has_transform_applied
            and path_element.pathGeometries[0].corner_radius is not None
        ):
            corner_radius = path_element.pathGeometries[0].corner_radius
            if any(corner_radius):  # only if there are values other than 0
                self.set_corner(path, corner_radius)

        # BaseElement
        path.label = path_element.name
        path.style["opacity"] = path_element.opacity
        path.style["mix-blend-mode"] = path_element.convert_blend()
        path.style["display"] = "none" if path_element.isHidden else "inline"
        if path_element.isLocked:
            path.set("sodipodi:insensitive", "true")

        # blur
        if path_element.blur > 0:
            self.set_blur(path, path_element.convert_blur())

        # Stroke Style
        if path_element.strokeStyle is not None:
            self.set_stroke_styles(path, path_element.strokeStyle)
        else:
            path.style["stroke"] = "none"

        # Fill Style
        if path_element.fillColor is not None:
            self.set_fill_color_styles(path, path_element.fillColor)
        elif path_element.fillGradient is not None:
            # Add gradientTransform
            # matrix transform is based on Vectornator 4.13.2, format 13
            # and Linearity Curve 5.1.1, format 21
            if path_element.fillGradient.transform is not None:
                path_element.fillGradient.gradient.set(
                    "gradientTransform", path_element.fillGradient.transform
                )

            elif (
                not self.has_transform_applied
                and path_element.localTransform is not None
            ):
                gradient_transform = path_element.localTransform.convert_transform()
                path_element.fillGradient.gradient.set(
                    "gradientTransform", gradient_transform
                )

            self.set_fill_grad_styles(path, path_element.fillGradient)
        else:
            path.style["fill"] = "none"

        return path

    def convert_text(self, text_element: VNTextElement) -> inkex.TextElement:
        """Converts a VNTextElement to an SVG Text (inkex.TextElement)."""
        text = inkex.TextElement()

        # transform
        if text_element.transform is not None:
            text.transform = text_element.transform
        elif not self.has_transform_applied and text_element.localTransform is not None:
            # remove scale to prevent over-compressed look
            text.transform = text_element.localTransform.convert_transform(
                with_scale=False
            )

        # BaseElement
        text.label = text_element.name
        text.style["opacity"] = text_element.opacity
        text.style["mix-blend-mode"] = text_element.convert_blend()
        text.style["display"] = "none" if text_element.isHidden else "inline"
        if text_element.isLocked:
            text.set("sodipodi:insensitive", "true")

        # blur
        if text_element.blur > 0:
            self.set_blur(text, text_element.convert_blur())

        if text_element.string and text_element.styledText:
            offset = 0  # global offset(letter count)
            y_offset = 0
            line_height = 1.0  # default paragraph margin

            paragraphs = text_element.string.split("\n")
            styled_index = 0  # current styledText
            remaining_length = 0  # remaining range of styledText

            for para in paragraphs:
                # line-tspan
                # TODO how should I implement align-center/left/right?
                line_tspan = inkex.Tspan()
                line_tspan.set("sodipodi:role", "line")
                line_tspan.set("x", "0")
                line_tspan.set("y", f"{y_offset}px")

                # inkex.utils.debug(f"this para: {repr(para)}")

                para_offset = 0  # offset in para
                while para_offset < len(para) + 1:
                    # break the loop if there's no remaining styledText
                    if styled_index >= len(text_element.styledText):
                        break

                    styled = text_element.styledText[styled_index]

                    # skip styledText which contains only "\n"
                    # ignore this check when offset == 0(start)
                    if offset != 0:
                        while (
                            styled.length == 1
                            and text_element.string[offset - 1] == "\n"
                        ):
                            offset += 1
                            para_offset += 1
                            styled_index += 1
                            # break the loop if there's no remaining styledText
                            if styled_index >= len(text_element.styledText):
                                break

                            styled = text_element.styledText[styled_index]

                    # set remaining_length
                    if remaining_length <= 0:
                        remaining_length = styled.length

                    # apply style "until" there's no more text in para
                    apply_length = min(len(para) + 1 - para_offset, remaining_length)
                    substring = para[para_offset : para_offset + apply_length]

                    tspan = inkex.Tspan()
                    tspan.text = substring
                    self.set_tspan_style(tspan, styled)

                    line_tspan.append(tspan)

                    # update offset
                    para_offset += apply_length
                    remaining_length -= apply_length

                    # move on to next styledText
                    if remaining_length <= 0:
                        styled_index += 1

                text.append(line_tspan)
                y_offset += (
                    styled.fontSize * line_height
                )  # last font decides line height
                offset += len(para) + 1

        return text

    def convert_base(self, base_element: VNBaseElement) -> inkex.PathElement:
        """Converts a VNBaseElement to an empty SVG path (inkex.PathElement)."""
        inkex.utils.debug(
            f"{base_element.name}: This element will be imported as empty path."
        )

        path = inkex.PathElement()

        # BaseElement
        path.label = base_element.name
        path.style["opacity"] = base_element.opacity
        path.style["mix-blend-mode"] = base_element.convert_blend()
        path.style["display"] = "none" if base_element.isHidden else "inline"
        if base_element.isLocked:
            path.set("sodipodi:insensitive", "true")

        # Style
        path.style["stroke"] = "none"
        path.style["fill"] = "none"

        return path

    def set_stroke_styles(
        self, elem: inkex.BaseElement, stroke: pathStrokeStyle
    ) -> None:
        """Apply pathStrokeStyle to inkex.BaseElement."""
        elem.style["stroke"] = stroke.color.hex
        elem.style["stroke-opacity"] = stroke.color.alpha
        elem.style["stroke-linecap"] = stroke.basicStrokeStyle.cap
        elem.style["stroke-linejoin"] = stroke.basicStrokeStyle.join
        elem.style["stroke-dasharray"] = stroke.basicStrokeStyle.dashPattern
        elem.style["stroke-width"] = stroke.width

    def set_fill_color_styles(self, elem: inkex.BaseElement, fill: VNColor) -> None:
        """Apply fillColor to inkex.BaseElement."""
        elem.style["fill"] = fill.hex
        elem.style["fill-opacity"] = fill.alpha
        elem.style["fill-rule"] = "nonzero"

    def set_fill_grad_styles(self, elem: inkex.BaseElement, fill: VNGradient) -> None:
        """Apply fillGradient to inkex.BaseElement."""
        self.document.defs.add(fill.gradient)
        elem.style["fill"] = f"url(#{fill.gradient.get_id()})"
        elem.style["fill-rule"] = "nonzero"

    def set_blur(
        self, elem: inkex.BaseElement, blur: inkex.Filter.GaussianBlur
    ) -> None:
        """Apply blur to inkex.BaseElement."""
        filt: inkex.Filter = inkex.Filter()
        filt.set("color-interpolation-filters", "sRGB")
        filt.add(blur)
        self.document.defs.add(filt)

        # Only one filter will be there
        elem.style["filter"] = f"url(#{filt.get_id()})"

    def set_corner(self, elem: inkex.PathElement, corner_radius: List[float]) -> None:
        """Apply rounded corner to inkex.PathElement."""
        params = " @ ".join(f"F,0,0,1,0,{r},0,1" for r in corner_radius)

        # FIXME more cornerRadius work
        #  flexible="false" is how Linearity Curve behaved,
        #  but this is not optimal
        path_effect = inkex.PathEffect.new(
            effect="fillet_chamfer",
            lpeversion="1",
            method="auto",
            flexible="false",
            is_visible="true",
            satellites_param=params,  # Inkscape 1.2
            nodesatellites_param=params,  # Inkscape 1.3
        )

        self.document.defs.add(path_effect)
        elem.set("inkscape:original-d", str(elem.path))
        elem.set("inkscape:path-effect", f"{path_effect.get_id(1)}")
        elem.attrib.pop("d", None)  # delete "d", Inkscape auto-generates LPE path

    def set_tspan_style(self, elem: inkex.Tspan, styled: singleStyledText) -> None:
        # weight, style
        # Condensed is not supported
        known_styles = {
            "Regular": ("normal", "normal"),
            "Plain": ("normal", "normal"),
            "Ultrajada": ("normal", "normal"),
            "Bold": ("bold", "normal"),
            "Italic": ("normal", "italic"),
            "Oblique": ("normal", "oblique"),
            "ExtraBold": ("bold", "normal"),  # ExtraBold as Bold
            "Light": ("300", "normal"),  # numbers may not be optimal
            "Medium": ("500", "normal"),
            "SemiBold": ("600", "normal"),
            "BoldOblique": ("bold", "oblique"),  # combined
            "BoldItalic": ("bold", "italic"),
        }

        font_parts = styled.fontName.split("-")

        # font name without styling
        base_font_parts = []
        font_weight = "normal"
        font_style = "normal"

        for part in font_parts:
            if part in known_styles:
                font_weight, font_style = known_styles[part]
            # name
            else:
                base_font_parts.append(part)

        base_font = " ".join(base_font_parts)
        base_font = f"'{base_font}'"

        elem.style["font-size"] = f"{styled.fontSize}px"
        elem.style["letter-spacing"] = f"{styled.kerning}px"
        elem.style["font-family"] = base_font
        elem.style["font-weight"] = font_weight
        elem.style["font-style"] = font_style

        # fill
        if styled.fillColor:
            self.set_fill_color_styles(elem, styled.fillColor)
        # elif styled.fillGradient:
        #    self.set_fill_grad_styles(tspan, styled.fillGradient)
        else:
            elem.style["fill"] = "none"

        ## stroke
        # if styled.strokeStyle:
        #    self.set_stroke_styles(tspan, styled.strokeStyle)

        ## decorations
        # if styled.underline:
        #    tspan.style["text-decoration"] = "underline"
        # if styled.strikethrough:
        #    tspan.style["text-decoration"] = "line-through"

    def add_guide(self, guide_element: VNBaseElement, offset: inkex.Vector2d) -> None:
        """
        Creates a Guide from VNGuideElement.

        Older Vectornator(4.10.4) uses actual path data for guides,
        which are converted into inkvn GroupElement.

        So argument type is set to VNBaseElement
        """

        def extract_guide(inkex_path: inkex.Path) -> Tuple[float, bool]:
            """
            Extracts guide offset and orientation from an inkex.Path object.
            """
            path_data = inkex_path.to_arrays()

            if len(path_data) < 2:
                raise ValueError("Invalid guide path: Less than two points found.")

            (_, [x0, y0]), (_, [x1, y1]) = path_data[0], path_data[-1]

            if y0 == y1:  # horizontal
                return y0, True
            elif x0 == x1:  # vertical
                return x0, False
            else:
                raise ValueError(
                    "Invalid guide path: Not perfectly horizontal or vertical."
                )

        orientation = False
        guide_offset = 0

        if isinstance(guide_element, VNGuideElement):
            guide_offset = guide_element.offset
            orientation = guide_element.orientation == 1

        elif isinstance(guide_element, VNGroupElement):
            # legacy guide element.
            # there should be two elements, second one is the line
            sub_element = guide_element.groupElements[1]
            if isinstance(sub_element, VNPathElement):
                path_data = sub_element.pathGeometries[0].path
                guide_offset, orientation = extract_guide(path_data)

        if orientation:
            self.document.namedview.add_guide(guide_offset + offset.y, orient=True)
        else:
            self.document.namedview.add_guide(guide_offset + offset.x, orient=False)



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005,2007 Aaron Spike, aaron@ekips.org
# - template dxf_outlines.dxf added Feb 2008 by Alvin Penner, penner@vaxxine.com
# - layers, transformation, flattening added April 2008 by Bob Cook, bob@bobcookdev.com
# - added support for dxf R12, Nov. 2008 by Switcher
# - brought together to replace ps2edit version 2018 by Martin Owens
# - added support for colors and refactored by Joshu Coats, joshu@fearchar.net
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from enum import Enum, IntEnum
from typing import IO, Union, Iterable

import inkex
import inkex.units
from inkex.bezier import cspsubdiv
from inkex.localization import inkex_gettext as _

from dxf_input import COLORS as DXF_COLORS

R12_HEADER = """ 0 
SECTION
 2 
HEADER
 9 
$ACADVER
 1 
AC1009
 9 
$EXTMIN
 10 
 0 
 20 
 0 
 9 
$EXTMAX
 10 
 8.5 
 20 
 11 
 0 
ENDSEC
 0 
SECTION
 2 
ENTITIES
"""

R12_FOOTER = """ 0 
ENDSEC
 0 
EOF"""


# DXF12 colors, derived from dxf_input.py, as (r,g,b) tuples
COLORS: dict[tuple[int, int, int], int] = {
    tuple(inkex.Color(hex_code)): index
    for index, hex_code in enumerate(DXF_COLORS)
    if hex_code != "PAD"
}


class Group(IntEnum):
    """DXF12 Group Codes"""

    ENTITY_START = 0
    LAYER_NAME = 8
    START_X = 10
    START_Y = 20
    COLOR = 62
    ENTITIES_FOLLOW = 66
    POLYLINE_FLAG = 70
    VERTEX_FLAG = 70


class PolylineFlag(IntEnum):
    none = 0
    closed = 1


class Entity(str, Enum):
    """DXF12 Entities"""

    POINT = "POINT"
    POLYLINE = "POLYLINE"
    VERTEX = "VERTEX"
    SEQEND = "SEQEND"

    def __str__(self) -> str:
        return self.value


def find_closest_color(
    color: tuple[int, int, int], colors: set[tuple[int, int, int]]
) -> tuple[int, int, int]:
    """Select the closest color from the set"""

    def _color_distance(c: tuple[int, int, int]) -> int:
        # No need to square-root the distance because only relative magnitude matters here
        return (c[0] - color[0]) ** 2 + (c[1] - color[1]) ** 2 + (c[2] - color[2]) ** 2

    sorted_colors = sorted(colors, key=_color_distance)
    return sorted_colors[0]


class DxfTwelve(inkex.OutputExtension):
    """
    Create dxf12 output from the svg

    DXF R12 reference is publicly available from AutoDesk:
    https://damassets.autodesk.net/content/dam/autodesk/www/developer-network/platform-technologies/autocad-dxf-archive/acad_r12_dxf.pdf
    """

    def __init__(self) -> None:
        super().__init__()
        self.color_mappings: dict[tuple[int, int, int], int] = {}
        self.handle: int = 255
        self.flatness: float = 0.1
        self._stream: IO

    def dxf_add(self, line: str) -> None:
        self._stream.write(line.encode("utf-8"))

    def dxf_insert_code(self, code: int, value: Union[str, float]) -> None:
        if isinstance(value, float):
            value = format(value, "f")
        self.dxf_add(f"{code}\n{value}\n")

    def dxf_start_entity(self, entity: Entity, layer: str) -> None:
        self.dxf_insert_code(Group.ENTITY_START, str(entity))
        self.dxf_insert_code(Group.LAYER_NAME, layer)

    def dxf_insert_color_code(self, color: inkex.Color) -> None:
        rgb_tuple = tuple(color.to_rgb())
        # If there's no color in the color_mappings dict, default to 256 ("inherit from layer")
        color_code = self.color_mappings.get(rgb_tuple, 256)
        self.dxf_insert_code(Group.COLOR, format(color_code, "d"))

    def dxf_point(
        self, layer: str, center: tuple[float, float], color: inkex.Color
    ) -> None:
        self.dxf_start_entity(Entity.POINT, layer)
        self.dxf_insert_color_code(color)
        self.dxf_insert_code(Group.START_X, center[0])
        self.dxf_insert_code(Group.START_Y, center[1])

    def dxf_polyline(
        self,
        layer: str,
        points: list[tuple[float, float]],
        color: inkex.Color,
        closed: bool,
    ) -> None:
        self.dxf_start_entity(Entity.POLYLINE, layer)
        self.dxf_insert_color_code(color)

        # Other values for the polyline flag are supported, per the docs provided by autodesk.net,
        # but we don't have them implemented here yet.
        polyline_flag = PolylineFlag.none
        if closed:
            polyline_flag |= PolylineFlag.closed
        self.dxf_insert_code(Group.POLYLINE_FLAG, polyline_flag)

        # DXF12's POLYLINEs start with a statement that X=0, Y=0. This is not because they all
        # start at the origin, but rather because in 3D mode, the Z value is used to give initial
        # elevation.
        self.dxf_insert_code(Group.START_X, 0.0)
        self.dxf_insert_code(Group.START_Y, 0.0)
        self.dxf_insert_code(Group.ENTITIES_FOLLOW, 1)
        for x, y in points:
            self.dxf_start_entity(Entity.VERTEX, layer)
            self.dxf_insert_code(Group.START_X, x)
            self.dxf_insert_code(Group.START_Y, y)
            self.dxf_insert_code(Group.VERTEX_FLAG, 0)
        self.dxf_start_entity(Entity.SEQEND, layer)

    def path_to_dxf_polyline(
        self, layer: str, path: inkex.CubicSuperPath, color: inkex.Color, closed: bool
    ) -> None:
        f = self.flatness
        is_flat = 0
        while is_flat < 1:
            try:
                # sub-divide paths in-place
                cspsubdiv(path, self.flatness)
                is_flat = 1
            except Exception:
                f += 0.1

        vertices: list[tuple[float, float]] = [
            vertex[1] for subpath in path for vertex in subpath
        ]
        self.dxf_polyline(layer, vertices, color, closed)

    def path_to_dxf_point(
        self, layer: str, path: inkex.CubicSuperPath, color: inkex.Color
    ) -> None:
        bbox = inkex.Path(path).bounding_box() or inkex.BoundingBox(0, 0)
        x, y = bbox.center
        self.dxf_point(layer, (x, y), color)

    def iter_relevant_nodes(
        self,
    ) -> Iterable[
        Union[
            inkex.PathElement,
            inkex.Rectangle,
            inkex.Circle,
            inkex.Polygon,
            inkex.Polyline,
            inkex.Ellipse,
            inkex.Line,
        ]
    ]:
        for node in self.svg.iterdescendants():
            if not isinstance(
                node,
                (
                    inkex.PathElement,
                    inkex.Rectangle,
                    inkex.Circle,
                    inkex.Polygon,
                    inkex.Polyline,
                    inkex.Ellipse,
                    inkex.Line,
                ),
            ):
                continue
            visible = True
            for parent in node.iterancestors():
                if isinstance(parent, (inkex.ClipPath, inkex.Mask, inkex.Defs)):
                    visible = False
                    break
            if not visible:
                continue
            yield node

    def populate_color_mappings(self, colors: set[tuple[int, int, int]]) -> None:
        """
        Using the set of colors used for paths in the SVG document, generate a mapping from those
        to the DXFv12-compatible colors.
        """

        if len(colors) > len(COLORS):
            raise inkex.AbortExtension(
                _(
                    "DXF12 supports a maximum of 249 distinct colors. Please simplify color "
                    "choices first."
                )
            )

        # Just in case there was something left over in here ...
        self.color_mappings.clear()

        # Also keep track of the colors we've already used
        color_codes_used: set[int] = set()

        # First, find all the colors that have an exact match
        for color in COLORS.keys() & colors:
            color_code = COLORS[color]
            self.color_mappings[color] = color_code
            color_codes_used.add(color_code)

        # Then choose colors from the remaining set for the unmatched ones
        remaining_colors = set(COLORS.keys() - self.color_mappings.keys())
        unmatched_colors = colors - COLORS.keys()
        for color in unmatched_colors:
            closest_color = find_closest_color(color, remaining_colors)
            color_code = COLORS[closest_color]
            self.color_mappings[color] = color_code
            color_codes_used.add(color_code)
            remaining_colors.discard(closest_color)

    def save(self, stream):
        # This extension only supports paths. Everything is converted to path.
        if len(self.svg.xpath("//svg:use|//svg:flowRoot|//svg:text")) > 0:
            self.preprocess(["flowRoot", "text"])

        self._stream = stream
        self.dxf_insert_code(999, '"DXF R12 Output" (www.mydxf.blogspot.com)')
        self.dxf_add(R12_HEADER)

        # Scale, but assume that the viewport is based on mm (dxf expects mm)
        scale = self.svg.scale / inkex.units.convert_unit("1mm", "px")
        h = self.svg.viewbox_height

        colors_used: set[tuple[int, int, int]] = set()

        for node in self.iter_relevant_nodes():
            node_style: inkex.Style = node.cascaded_style()
            colors_used.add(tuple(node_style.get_color("stroke").to_rgb()))

        self.populate_color_mappings(colors_used)

        for node in self.iter_relevant_nodes():
            # TODO: this assumes that all elements are direct descendants of layers
            layer = node.getparent().label

            if layer is None:
                layer = "Layer 1"

            node.transform = node.composed_transform()
            node.transform = (
                inkex.Transform(((scale, 0, 0), (0, -scale, h * scale)))
                @ node.transform
            )
            node_style: inkex.Style = node.cascaded_style()
            color: inkex.Color = node_style.get_color("stroke")

            transformed_path = node.path.transform(node.transform)

            if layer.lower().endswith("drill"):
                self.path_to_dxf_point(layer, transformed_path.to_superpath(), color)
                continue

            for subpath in transformed_path.break_apart():
                closed = any(isinstance(pe, inkex.paths.ZoneClose) for pe in subpath)
                self.path_to_dxf_polyline(layer, subpath.to_superpath(), color, closed)

        self.dxf_add(R12_FOOTER)


if __name__ == "__main__":
    DxfTwelve().run()



"""
VNBaseElement, VNTransform
"""

from __future__ import annotations

import math
from dataclasses import dataclass, field
from typing import List, Optional

import inkex


@dataclass
class VNBaseElement:
    """Common Element properties."""

    name: str
    blur: float
    opacity: float
    blendMode: int
    isHidden: bool
    isLocked: bool
    localTransform: Optional[VNTransform]

    def convert_blend(self) -> str:
        """Returns value for mix-blend-mode attribute."""
        blend_mode_map = {
            0: "normal",
            1: "multiply",
            2: "screen",
            3: "overlay",
            4: "darken",
            5: "lighten",
            10: "difference",
            11: "exclusion",
            12: "hue",
            13: "saturation",
            14: "color",
            15: "luminosity",
        }
        return blend_mode_map.get(self.blendMode, "normal")

    def convert_blur(self) -> inkex.Filter.GaussianBlur:
        """Returns inkex GaussianBlur."""
        """
        FIXME Blur does not work like it should.
        divided by 3 because I don't know
        """
        return inkex.Filter.GaussianBlur.new(
            stdDeviation=self.blur / 3.0, result="blur"
        )


@dataclass
class VNTransform:
    """Linearity Curve transform."""

    rotation: float = 0.0
    scale: List[float] = field(default_factory=lambda: [1.0, 1.0])
    shear: float = 0.0
    translation: List[float] = field(default_factory=lambda: [0.0, 0.0])

    def convert_transform(
        self, keep_proportion=False, with_scale=True
    ) -> inkex.transforms.Transform:
        """
        Creates a transform string for the `g` element in inkex.transforms.Transform.

        keep_proportion applies scaling in x axis to y axis as well.
        with_scale determines whether to include scale or not.
        """

        # Extract values
        rotation_deg = math.degrees(self.rotation)
        sx, sy = self.scale
        # Shear is given in radians
        shear_deg = math.degrees(math.atan(self.shear))
        tx, ty = self.translation

        # Create transform components
        # The order matters
        tr = inkex.transforms.Transform()
        if tx != 0 or ty != 0:
            # Translate by (tx, ty)
            tr.add_translate(tx, ty)

        if self.rotation != 0:
            # Rotate around origin (adjust if a specific pivot is needed)
            tr.add_rotate(rotation_deg)

        if with_scale and (sx != 1 or sy != 1):
            # Scale by (sx, sy)
            if keep_proportion == True:
                tr.add_scale(sx)
            else:
                tr.add_scale(sx, sy)

        if self.shear != 0:
            # Skew in the X direction
            tr.add_skewx(shear_deg)

        return tr



# coding=utf-8
#
# Copyright (C) 2010 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
Some basic common code shared between EAN and UCP generators.
"""

from .Base import Barcode, TEXT_POS_TOP
from inkex.localization import inkex_gettext as _

try:
    from typing import Optional, List, Dict
except ImportError:
    pass

MAPPING = [
    # Left side of barcode Family '0'
    [
        "0001101",
        "0011001",
        "0010011",
        "0111101",
        "0100011",
        "0110001",
        "0101111",
        "0111011",
        "0110111",
        "0001011",
    ],
    # Left side of barcode Family '1' and flipped to right side.
    [
        "0100111",
        "0110011",
        "0011011",
        "0100001",
        "0011101",
        "0111001",
        "0000101",
        "0010001",
        "0001001",
        "0010111",
    ],
]
# This chooses which of the two encodings above to use.
FAMILIES = (
    "000000",
    "001011",
    "001101",
    "001110",
    "010011",
    "011001",
    "011100",
    "010101",
    "010110",
    "011010",
)


class EanBarcode(Barcode):
    """Simple base class for all EAN type barcodes"""

    lengths = None  # type: Optional[List[int]]
    length = None  # type: Optional[int]
    checks = []  # type: List[int]
    extras = {}  # type: Dict[int, str]
    magic = 10
    guard_bar = "202"
    center_bar = "02020"

    @staticmethod
    def intarray(number):
        """Convert a string of digits into an array of ints"""
        return [int(i) for i in number]

    @staticmethod
    def encode_interleaved(family, number, fams=FAMILIES):
        """Encode any side of the barcode, interleaved"""
        result = []
        encset = EanBarcode.intarray(fams[family])
        for i, _ in enumerate(number):
            thismap = MAPPING[encset[i]]
            result.append(thismap[number[i]])
        return result

    @staticmethod
    def encode_right(number):
        """Encode the right side of the barcode, non-interleaved"""
        result = []
        for num in number:
            # The right side is always the reverse of the left's family '1'
            result.append(MAPPING[1][num][::-1])
        return result

    @staticmethod
    def encode_left(number):
        """Encode the left side of the barcode, non-interleaved"""
        result = []
        for num in number:
            result.append(MAPPING[0][num])
        return result

    @staticmethod
    def space(*spacing):
        """Space out an array of numbers"""
        result = ""
        for space in spacing:
            if isinstance(space, list):
                for i in space:
                    result += str(i)
            elif isinstance(space, int):
                result += " " * space
        return result

    def get_lengths(self):
        """Return a list of acceptable lengths"""
        if self.length:
            return [self.length]
        return self.lengths[:]

    def encode(self, text):
        """Encode any EAN barcode"""
        code = text.replace(" ", "").strip()
        guide = code.endswith(">")
        code = code.strip(">")

        if not code.isdigit():
            return self.error(code, _("Not a Number, must be digits 0-9 only"))
        lengths = self.get_lengths() + self.checks

        # Allow extra barcodes after the first one
        if len(code) not in lengths:
            for extra in self.extras:
                sep = len(code) - extra
                if sep in lengths:
                    # Generate a barcode along side this one.
                    self.add_extra_barcode(
                        self.extras[extra],
                        text=code[sep:],
                        x=self.pos_x + 400 * self.scale,
                        text_pos=TEXT_POS_TOP,
                    )
                    code = code[:sep]

        if len(code) not in lengths:
            return self.error(
                code,
                _("Wrong size {:d}, must be {} digits").format(
                    len(code), ", ".join([str(length) for length in lengths])
                ),
            )

        if self.checks:
            if len(code) not in self.checks:
                code = self.append_checksum(code)
            elif not self.verify_checksum(code):
                return self.error(code, _("Checksum failed, omit for new sum"))
        return self._encode(EanBarcode.intarray(code), guide=guide)

    def _encode(self, num, guide=False):
        """
        Write your EAN encoding function, it's passed in an array of int and
        it should return a string on 1 and 0 for black and white parts
        """
        raise NotImplementedError("_encode should be provided by parent EAN")

    def enclose(self, left, right=()):
        """Standard Enclosure"""
        parts = [self.guard_bar] + left
        parts.append(self.center_bar)
        parts += list(right) + [self.guard_bar]
        return "".join(parts)

    def get_checksum(self, num):
        """Generate a UPCA/EAN13/EAN8 Checksum"""
        # Left to right,checksum based on first digits.
        total = sum([int(n) * (3, 1)[x % 2] for x, n in enumerate(num[::-1])])
        # Modulous result to a single digit checksum
        checksum = self.magic - (total % self.magic)
        if checksum < 0 or checksum >= self.magic:
            return "0"
        return str(checksum)

    def append_checksum(self, number):
        """Apply the checksum to a short number"""
        return number + self.get_checksum(number)

    def verify_checksum(self, number):
        """Verify any checksum"""
        return self.get_checksum(number[:-1]) == number[-1]



"""
VNGuideElement
"""

from dataclasses import dataclass

from .base import VNBaseElement


@dataclass
class VNGuideElement(VNBaseElement):
    """Guide element."""

    offset: float
    orientation: int  # 0: vertical, 1: horizontal



#
# Copyright 2022 Simon Duerr dev@simonduerr.eu
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
Access Bioicons.

Downloads the database to cache and uses it locally for searching.
"""

import json
import requests
from datetime import datetime
import os


from cachecontrol import CacheControl, CacheControlAdapter
from cachecontrol.caches.file_cache import FileCache
from cachecontrol.heuristics import ExpiresAfter

from import_sources import RemoteSource


def local_search(query, db):
    for item in db:
        if query.lower() in item["name"].lower():
            yield item
    return


class Bioicons(RemoteSource):
    name = "Bioicons"
    icon = "sources/bioicons.svg"
    db_url = "https://bioicons.com/icons/icons.json"
    icon_url = "https://bioicons.com/icons/"

    def __init__(self, cache_dir):
        self.session = requests.session()
        self.cache_dir = cache_dir
        self.session.mount(
            "https://",
            CacheControlAdapter(
                cache=FileCache(cache_dir),
                heuristic=ExpiresAfter(days=5),
            ),
        )
        # check if local db is up to date with Last Modified header
        try:
            response = requests.head(self.db_url)
        except Exception:
            response = None

        self._json = os.path.join(self.cache_dir, "icons.json")
        if not os.path.isfile(self._json):
            last_modified = 0
        else:
            last_modified = os.path.getmtime(self._json)

        if response is not None:
            last_update = datetime.strptime(
                response.headers["Last-Modified"], "%a, %d %b %Y %H:%M:%S GMT"
            ).timestamp()
            # if icon db was modified download new database
            if last_modified < last_update:
                self.to_local_file(self.db_url)

    def search(self, query):
        results = []
        if os.path.isfile(self._json):
            with open(self._json, "r") as f:
                db = json.load(f)
            results = local_search(query, db)

        for item in results:
            yield {
                "id": item["name"],
                "name": item["name"],
                "summary": item["category"],
                "created": None,
                "popularity": 0,
                "author": item["author"],
                "thumbnail": f"{self.icon_url}{item['license']}/{item['category']}/{item['author']}/{item['name']}.svg",
                "file": f"{self.icon_url}{item['license']}/{item['category']}/{item['author']}/{item['name']}.svg",
                "license": item["license"],
            }
